import json
from typing import List

import pytest  # type: ignore
import responses  # type: ignore

from pytos2.securechange import ScwAPI
from pytos2.securechange.entrypoint import Scw
from pytos2.securechange.ticket import Ticket, ApplicationDetails
from tests.securetrack.conftest import (
    st,
    network_objects_mock,
    devices_mock,
    device_rules_mock,
)  # noqa
from pytos2.securechange.fields import MultiAccessRequest

from pytos2.securetrack.device import Device
from pytos2.securetrack.rule import SecurityRule
from pytos2.securetrack.network_object import NetworkObject

from pytos2.securetrack import St


class MockExit(Exception):
    pass


def _api_path(path):
    return f"https://198.18.0.1/securechangeworkflow/api/securechange/{path}"


@pytest.fixture
def st_api():
    return StAPI(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def st():
    return St(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def device_20_network_objects():
    js = json.load(open("tests/securetrack/json/network_objects/20.json"))
    objects = [
        NetworkObject.kwargify(j) for j in js["network_objects"]["network_object"]
    ]
    return objects


@pytest.fixture
def device_20_rules():
    dev = json.load(open("tests/securetrack/json/devices/device-20.json"))
    dev = Device.kwargify(dev)

    js = json.load(open("tests/securetrack/json/rules/device-20.json"))
    rules = [SecurityRule.kwargify(j) for j in js["rules"]["rule"]]

    for rule in rules:
        rule.device = dev

    return rules


@pytest.fixture
def mock_devices() -> List[Device]:
    js = json.load(open("tests/securetrack/json/devices/devices.json"))
    devices = [Device.kwargify(j) for j in js["devices"]["device"]]
    return devices


@pytest.fixture
def st_devices_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices",
        json=json.load(open(f"tests/securetrack/json/devices/devices.json")),
    )


@pytest.fixture
def device_20_rules_mock(st, st_devices_mock):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/20/rules",
        json=json.load(open(f"tests/securetrack/json/rules/device-20.json")),
    )


@pytest.fixture
def device_1_rules():
    dev = json.load(open("tests/securetrack/json/devices/device-1.json"))
    dev = Device.kwargify(dev)

    js = json.load(open("tests/securetrack/json/rules/device-1.json"))
    rules = [SecurityRule.kwargify(j) for j in js["rules"]["rule"]]

    for rule in rules:
        rule.device = dev

    return rules


@pytest.fixture
def all_fields_ticket_json():
    return json.load(open("tests/securechange/json/ticket/all_fields-ticket.json"))


@pytest.fixture
def all_fields_workflow_json(scw_api, all_fields_ticket_json):
    return json.load(open("tests/securechange/json/all_fields-workflow.json"))


@pytest.fixture
def all_fields_step_json(all_fields_ticket_json):
    return [step["name"] for step in all_fields_ticket_json["ticket"]["steps"]["step"]]


@pytest.fixture
def all_fields_field_json(all_fields_ticket_json):
    return all_fields_ticket_json["ticket"]["steps"]["step"][0]["tasks"]["task"][
        "fields"
    ]


@pytest.fixture
def application_details():
    return ApplicationDetails.kwargify(
        json.load(open("tests/securechange/json/application_details.json"))
    )


@pytest.fixture
def post_bad_ticket_json():
    json.load(open("tests/securechange/json/ticket/post_bad_ticket.json"))


@pytest.fixture
def closed_group_modify_ticket(scw_api):
    return Ticket.kwargify(
        json.load(
            open("tests/securechange/json/ticket/closed_group_modify-ticket.json")
        )
    )


@pytest.fixture
def closed_ticket(scw_api):
    return Ticket.kwargify(json.load(open("tests/securechange/json/ticket/redo.json")))


@pytest.fixture
def first_step_ticket(scw_api):
    ticket = json.load(open("tests/securechange/json/ticket/redo.json"))
    redo = ticket["ticket"]
    step_one = redo["steps"]["step"][0]
    redo["status"] = "In Progress"
    redo["steps"]["step"] = [step_one]
    redo["current_step"] = {"id": step_one["id"], "name": step_one["name"]}
    return Ticket.kwargify(ticket)


@pytest.fixture
def redo_ticket(scw_api):
    ticket = json.load(open("tests/securechange/json/ticket/redo.json"))
    redo = ticket["ticket"]
    step_two = redo["steps"]["step"][1]
    redo["status"] = "In Progress"
    redo["steps"]["step"] = redo["steps"]["step"][0:2]
    redo["current_step"] = {"id": step_two["id"], "name": step_two["name"]}
    return Ticket.kwargify(ticket)


@pytest.fixture
def open_access_request_ticket(scw_api):
    return Ticket.kwargify(
        json.load(
            open("tests/securechange/json/ticket/open_with_access_request-ticket.json")
        )
    )


@pytest.fixture
def open_group_modify_ticket(scw_api):
    return Ticket.kwargify(
        json.load(
            open("tests/securechange/json/ticket/open_with_group_modify-ticket.json")
        )
    )


@pytest.fixture
def ar_with_designer_results():
    return MultiAccessRequest.kwargify(
        json.load(
            open(
                "tests/securechange/json/field/multi_access_request_with_designer_results-field.json"
            )
        )
    )


@pytest.fixture
def tickets_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets",
        json=json.load(open("tests/securechange/json/ticket/tickets-0-99.json")),
    )

    responses.add(
        responses.GET,
        "https://10.100.0.1/securechangeworkflow/api/securechange/tickets/fake-url-due-to-params/count_100/start_100",
        json=json.load(open("tests/securechange/json/ticket/tickets-100-199.json")),
    )


@pytest.fixture
def scw_api():
    return ScwAPI(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def scw():
    return Scw(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def ticket_search_mock():
    ticket_search_by_status = json.load(
        open("tests/securechange/json/ticket/ticket_search_assigned.json")
    )

    ticket_search_by_step = json.load(
        open("tests/securechange/json/ticket/ticket_search_by_step_name.json")
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search?status=ASSIGNED",
        json=ticket_search_by_status,
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/search?current_step=manual%20step",
        json=ticket_search_by_step,
    )


@pytest.fixture
@responses.activate
def all_fields_mock(scw_api, all_fields_ticket_json):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/283",
        json=all_fields_ticket_json,
    )
    return Ticket.get(283)


@pytest.fixture
def all_fields_ticket(scw_api, all_fields_ticket_json):
    return Ticket.kwargify(all_fields_ticket_json)


@pytest.fixture
def dynamic_assignment_ticket(scw_api, all_fields_ticket_json):
    return Ticket.kwargify(
        json.load(open("tests/securechange/json/ticket/dynamic_assignment-ticket.json"))
    )


@pytest.fixture
def get_test_field():
    def f(field_cls):
        xsi_type = [
            a.default for a in field_cls.__attrs_attrs__ if a.name == "xsi_type"
        ][0]
        return field_cls.kwargify(
            json.load(
                open(f"tests/securechange/json/field/{xsi_type.value}-field.json")
            )
        )

    return f


@pytest.fixture
def users_mock():
    users_json = json.load(open("tests/securechange/json/users/users_by_username.json"))
    all_users_json = json.load(open("tests/securechange/json/users/all_users.json"))

    responses.add(
        responses.GET,
        _api_path("users?user_name=r&exact_name=True"),
        match_querystring=True,
        json=users_json,
    )

    responses.add(responses.GET, _api_path("users"), json=all_users_json)


@pytest.fixture
def user_mock():
    user_json = json.load(open("tests/securechange/json/users/user_45.json"))

    responses.add(
        responses.GET, _api_path("users/45"), match_querystring=True, json=user_json
    )

    pass



import pytest
import json
import responses

from pytos2.securetrack.join_cloud import JoinCloud
from pytos2.utils import get_api_node


class TestJoinCloud:
    cloud = json.load(open("tests/securetrack/json/join_clouds/cloud-67.json"))

    @responses.activate
    def test_add_join_cloud_200(self, st):
        """Add cloud"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=201,
        )

        """Add single cloud"""
        newCloud = st.add_join_cloud(self.cloud)
        assert newCloud == None

    @responses.activate
    def test_add_join_cloud_400(self, st):
        """Add cloud"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=400,
        )

        """Add single cloud"""
        with pytest.raises(ValueError) as exception:
            st.add_join_cloud(self.cloud)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_join_cloud_404(self, st):
        """Add cloud"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=404,
        )

        """Add single cloud"""
        with pytest.raises(ValueError) as exception:
            st.add_join_cloud(self.cloud)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_join_cloud(self, st, join_cloud_mock):
        """Get cloud by device id"""
        join_cloud = JoinCloud.kwargify(self.cloud)
        cloudByInt = st.get_join_cloud(67)
        cloudByStr = st.get_join_cloud("67")
        assert cloudByInt == join_cloud
        assert cloudByStr == join_cloud

        with pytest.raises(ValueError) as exception:
            st.get_join_cloud(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_join_cloud("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_join_cloud_200(self, st):
        """Update cloud"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=200,
        )

        """Update single cloud"""
        newCloud = st.update_join_cloud(self.cloud)
        assert newCloud == None

    @responses.activate
    def test_update_join_cloud_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            st.update_join_cloud(self.cloud)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_join_cloud_404(self, st):
        """PUT device id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            st.update_join_cloud(self.cloud)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_join_cloud(self, st, join_cloud_mock):
        """Delete cloud by cloud id"""
        cloudByInt = st.delete_join_cloud(67)
        cloudByStr = st.delete_join_cloud("67")
        assert cloudByInt == None
        assert cloudByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_join_cloud(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_join_cloud("404")
        assert "Not Found" in str(exception.value)



from typing import Union, Optional, List

from result import Result
from requests import Response
from requests.exceptions import HTTPError

# avoid circular imports
import pytos2  # noqa
from .api import SaAPI
from pytos2.utils import NoInstance, get_api_node
from .application_identities import ApplicationIdentity


class Sa:
    default: Union["Sa", NoInstance] = NoInstance(
        "Sa.default",
        "No Sa instance has been initialized yet, initialize with `Sa(*args, **kwargs)`",
    )

    def __init__(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        default=True,
    ):
        self.api: SaAPI = SaAPI(hostname, username, password)
        if default:
            Sa.default = self

        self._application_identities: List[ApplicationIdentity] = []

    @property
    def application_identities(self) -> List[ApplicationIdentity]:
        if not self._application_identities:
            res = self._get_application_identities()
            self._application_identities = res
        return self._application_identities

    def _get_application_identities(
        self, cache=True
    ) -> Result[List[ApplicationIdentity], Response]:
        res = self.api.get_application_identities()
        if res.ok:
            return [
                ApplicationIdentity.kwargify(a)
                for a in get_api_node(
                    res.json(),
                    "application_identities.application_identity",
                    listify=True,
                )
            ]
        else:
            try:
                msg = res.text
                res.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"got '{msg}' from API Error :{e}")



from enum import Enum

from typing import Optional, List, Union
from netaddr import IPAddress
from attr.converters import optional

from pytos2.utils import propify, prop
from pytos2.models import Jsonable

from pytos2.securechange.rule import SlimRule
from pytos2.securechange.designer import DesignerResults
from pytos2.securechange.designer_verifier_common import NetworkObject

from pytos2.securetrack import St
from pytos2.securetrack.service_object import Service as STService

from pytos2.securetrack.rule import SecurityRule
from pytos2.securetrack.network_object import (
    NetworkObjectXsiType as STNetworkObjectXsiType,
    NetworkObject as STNetworkObject,
)

from .binding import Binding
from . import Designer, Field, FieldXsiType


def _mv_key(j, src, dest):
    if not src in j:
        return

    j[dest] = j[src]
    del j[src]


def _optional_str(o):
    if o is None:
        return None
    else:
        return str(o)


@propify
class ModificationNetworkObject(Jsonable):
    class XsiType(Enum):
        NETWORK_OBJECT = "ns_sc_policy:network_object"
        SUBNET_NETWORK_OBJECT = "ns_sc_policy:subnet_network_object"
        HOST_NETWORK_OBJECT = "ns_sc_policy:host_network_object"
        RANGE_NETWORK_OBJECT = "ns_sc_policy:range_network_object"
        NETWORK_OBJECT_GROUP = "ns_sc_policy:network_object_group"
        HOST_NETWORK_OBJECT_WITH_INTERFACES = (
            "ns_sc_policy:host_network_object_with_interfaces"
        )

    xsi_type: XsiType = prop(XsiType.NETWORK_OBJECT)
    id: Optional[int] = prop(None)
    comment: Optional[str] = prop(None)
    name: Optional[str] = prop(None)

    def from_securetrack(obj):
        j = dict(obj.data or obj._json)
        if obj.xsi_type == STNetworkObjectXsiType.RANGE_NETWORK_OBJECT:
            _mv_key(j, "first_ip", "min_ip")
            _mv_key(j, "last_ip", "max_ip")

        cls = {
            STNetworkObjectXsiType.HOST_NETWORK_OBJECT: ModificationHostNetworkObject,
            STNetworkObjectXsiType.HOST_NETWORK_OBJECT_WITH_INTERFACES: ModificationHostNetworkObjectWithInterfaces,
            STNetworkObjectXsiType.NETWORK_OBJECT_GROUP: ModificationNetworkObjectGroup,
            STNetworkObjectXsiType.RANGE_NETWORK_OBJECT: ModificationRangeNetworkObject,
            STNetworkObjectXsiType.SUBNET_NETWORK_OBJECT: ModificationSubnetNetworkObject,
        }.get(obj.xsi_type, ModificationNetworkObject)
        return cls.kwargify(j)


@propify
class ModificationHostNetworkObject(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.HOST_NETWORK_OBJECT
    )
    netmask: Optional[IPAddress] = prop(
        IPAddress("255.255.255.255"),
        converter=optional(IPAddress),
        jsonify=_optional_str,
    )
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=_optional_str
    )


@propify
class ModificationSubnetNetworkObject(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.SUBNET_NETWORK_OBJECT
    )
    netmask: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=_optional_str
    )
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=_optional_str
    )


@propify
class ModificationRangeNetworkObject(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.RANGE_NETWORK_OBJECT
    )
    min_ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=_optional_str
    )
    max_ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=_optional_str
    )


@propify
class ModificationNetworkObjectGroup(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.NETWORK_OBJECT_GROUP
    )


@propify
class ModificationHostNetworkObjectWithInterfaces(ModificationNetworkObject):
    xsi_type: ModificationNetworkObject.XsiType = prop(
        ModificationNetworkObject.XsiType.HOST_NETWORK_OBJECT_WITH_INTERFACES
    )
    netmask: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=_optional_str
    )
    ip: Optional[IPAddress] = prop(
        None, converter=optional(IPAddress), jsonify=_optional_str
    )


@propify
class RangeDTO(Jsonable):
    class Prop(Enum):
        FROM = "from"

    to: int = prop()
    from_: int = prop(key=Prop.FROM.value)


@propify
class ModificationServiceObject(Jsonable):
    class XsiType(Enum):
        SERVICE_OBJECT = "ns_sc_policy:service_object"
        ICMP_SERVICE = "ns_sc_policy:icmp_service"
        IP_SERVICE = "ns_sc_policy:ip_service"
        SERVICE_GROUP = "ns_sc_policy:service_group"
        TRANSPORT_SERVICE = "ns_sc_policy:transport_service"

    xsi_type: XsiType = prop(XsiType.SERVICE_OBJECT, key=Jsonable.Prop.XSI_TYPE.value)

    id: Optional[int] = prop(None)
    comment: Optional[str] = prop(None)
    name: Optional[str] = prop(None)

    def from_securetrack(obj):
        from pytos2.securetrack.service_object import (
            TCPServiceObject,
            UDPServiceObject,
            ICMPServiceObject,
            ServiceGroup,
        )

        if isinstance(obj, (TCPServiceObject, UDPServiceObject)):
            protos = {6: "TCP", 17: "UDP"}

            proto = protos.get(obj.protocol, obj.protocol)

            return ModificationTransportService(
                comment=obj.comment,
                name=obj.name,
                protocol=proto,
                port=RangeDTO(from_=obj.min_port, to=obj.max_port),
            )
        elif isinstance(obj, (ICMPServiceObject)):
            return ModificationIcmpService(
                comment=obj.comment,
                name=obj.name,
                type=RangeDTO(from_=obj.min_port, to=obj.max_port),
            )
        elif isinstance(obj, (ServiceGroup)):
            return ModificationServiceGroup(comment=obj.comment, name=obj.name)
        else:
            return ModificationServiceObject(comment=obj.comment, name=obj.name)


@propify
class ModificationIcmpService(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.ICMP_SERVICE, key=Jsonable.Prop.XSI_TYPE.value
    )

    type: Optional[RangeDTO] = prop(None)


@propify
class ModificationIPService(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.IP_SERVICE, key=Jsonable.Prop.XSI_TYPE.value
    )

    protocol: Optional[RangeDTO] = prop(None)


@propify
class ModificationServiceGroup(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.SERVICE_GROUP,
        key=Jsonable.Prop.XSI_TYPE.value,
    )


@propify
class ModificationTransportService(ModificationServiceObject):
    xsi_type: ModificationServiceObject.XsiType = prop(
        ModificationServiceObject.XsiType.TRANSPORT_SERVICE,
        key=Jsonable.Prop.XSI_TYPE.value,
    )

    protocol: Optional[str] = prop(None)
    port: Optional[RangeDTO] = prop(None)


@propify
class RuleOperationBinding(Jsonable):
    binding_uid: str = prop()
    binding: Optional[Binding] = prop(None)
    rules: List[SlimRule] = prop(factory=list, flatify="rule")


@propify
class RuleOperationDevice(Jsonable):
    revision_id: Optional[int] = prop(None)
    management_id: int = prop()

    management_name: str = prop("")
    management_ip: Optional[IPAddress] = prop(None, converter=optional(IPAddress))

    revision_number: Optional[int] = prop(None)
    number_of_rules: Optional[int] = prop(None)

    bindings: List[RuleOperationBinding] = prop(factory=list, flatify="binding")


@propify
class RuleOperation(Field):
    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_RECERTIFICATION, key=Jsonable.Prop.XSI_TYPE.value
    )

    devices: List[RuleOperationDevice] = prop(factory=list, flatify="device")

    def get_rules(self):
        rules = []

        for device in self.devices:
            for binding in device.bindings:
                rules += binding.rules

        return rules

    def is_rule_added(self, rule: SecurityRule):
        rules = self.get_rules()
        for slim in rules:
            if slim.uid == rule.uid:
                return True

        return False

    def add_rule(self, rule: SecurityRule):
        if len(self.devices) > 0:
            device = self.devices[0]
            if (
                device.management_id != rule.device.id
                or device.revision_id != rule.device.latest_revision
            ):
                return None
        else:
            device = RuleOperationDevice(
                management_id=rule.device.id, revision_id=rule.device.latest_revision
            )
            self.devices.append(device)

        binding = None
        if len(device.bindings) > 0:
            for _b in device.bindings:
                if _b.binding_uid == rule.bindings[0].uid:
                    binding = _b
                    break

        if binding is None:
            binding = RuleOperationBinding(binding_uid=rule.bindings[0].uid)
            device.bindings.append(binding)

        slim_rule = SlimRule.kwargify({"uid": rule.uid})

        binding.rules.append(slim_rule)


@propify
class ObjectStatus(Jsonable):
    class XsiType(Enum):
        EXISTING = "existing"
        NEW = "new"

    xsi_type: XsiType = prop(XsiType.EXISTING, key=Jsonable.Prop.XSI_TYPE.value)
    st_uid: Optional[str] = prop(None)


@propify
class DeviceNetworkObject(Jsonable):
    network_object: List[NetworkObject] = prop(factory=list)
    device_id: Optional[int] = prop(None)
    status: Optional[ObjectStatus] = prop(None)


@propify
class DeviceServiceObject(Jsonable):
    service_object: List[ModificationServiceObject] = prop(factory=list)
    uid: Optional[str] = prop(None)
    device_id: Optional[int] = prop(None)
    status: Optional[ObjectStatus] = prop(None)
    id: Optional[int] = prop(None)


class CellAction(Enum):
    REMOVE = "REMOVE"
    ADD = "ADD"


@propify
class NetworkObjectCellModification(Jsonable):
    device_network_object: DeviceNetworkObject = prop()
    action: Optional[CellAction] = prop(None)
    id: Optional[int] = prop(None)


@propify
class NetworkObjectCellModifications(Jsonable):
    id: Optional[int] = prop(None)
    network_object_cell_modifications: List[NetworkObjectCellModification] = prop(
        factory=list, flatify="network_object_cell_modification"
    )


@propify
class ServiceObjectCellModification(Jsonable):
    device_service_object: DeviceServiceObject = prop()
    action: Optional[CellAction] = prop(None)
    id: Optional[int] = prop(None)


@propify
class ServiceObjectCellModifications(Jsonable):
    id: Optional[int] = prop(None)
    service_object_cell_modifications: List[ServiceObjectCellModification] = prop(
        factory=list, flatify="service_object_cell_modification"
    )


@propify
class RuleKey(Jsonable):
    device_id: Optional[int] = prop(None)
    binding_uid: Optional[str] = prop(None)
    rule_uid: Optional[str] = prop(None)

    def __eq__(a, b):
        return all(
            [
                a.device_id == b.device_id,
                a.binding_uid == b.binding_uid,
                a.rule_uid == b.rule_uid,
            ]
        )


@propify
class RuleModification(Jsonable):
    class XsiType(Enum):
        RULE_MODIFICATION = "rule_modification"
        MODIFY_RULE_MODIFICATION = "modify_rule_modification"

    rule_key: RuleKey = prop()
    id: Optional[int] = prop(None)
    xsi_type: XsiType = prop(
        XsiType.RULE_MODIFICATION, key=Jsonable.Prop.XSI_TYPE.value
    )

    _rule: SecurityRule = prop(None, jsonify=False)

    @property
    def rule(self):
        if self._rule:
            return self._rule

        rules = St.default.get_rules(
            device=self.rule_key.device_id, rule_uid=self.rule_key.rule_uid
        )
        rule = rules[0]
        self._rule = rule
        return self._rule


@propify
class ModifyRuleModification(RuleModification):
    xsi_type: RuleModification.XsiType = prop(
        RuleModification.XsiType.MODIFY_RULE_MODIFICATION,
        key=Jsonable.Prop.XSI_TYPE.value,
    )
    source_modifications: Optional[NetworkObjectCellModifications] = prop(None)
    destination_modifications: Optional[NetworkObjectCellModifications] = prop(None)
    service_modifications: Optional[ServiceObjectCellModifications] = prop(None)

    def _add_service_obj_mod(self, modlist, objlist, obj, device_id, cell_action):
        if isinstance(obj, str):
            obj = St.default.get_service(obj, device_id)
            _uid = obj.uid

            obj = ModificationServiceObject.from_securetrack(obj)
        elif isinstance(obj, STService):
            _uid = obj.uid

            obj = ModificationServiceObject.from_securetrack(obj)
        else:
            _uid = None

        if not obj:
            return False

        obj_status = ObjectStatus(
            xsi_type=ObjectStatus.XsiType.EXISTING if _uid else ObjectStatus.XsiType.NEW
        )
        if _uid:
            obj_status.st_uid = _uid

        svc_mod = ServiceObjectCellModification(
            action=cell_action,
            device_service_object=DeviceServiceObject(
                device_id=device_id, status=obj_status
            ),
        )

        found = False
        for ref in objlist:
            if obj_status.st_uid and ref.uid == obj_status.st_uid:
                found = True
                break

        if (cell_action == CellAction.REMOVE and found) or (
            cell_action == CellAction.ADD and not found
        ):
            svc_mod.device_service_object.service_object.append(obj)
            modlist.append(svc_mod)
            return True
        else:
            return False

    def _add_net_obj_mod(self, modlist, objlist, obj, device_id, cell_action):
        if isinstance(obj, str):
            obj = St.default.get_network_object(obj, device_id)
            _uid = obj.uid
            obj = ModificationNetworkObject.from_securetrack(obj)
        elif isinstance(obj, STNetworkObject):
            _uid = obj.uid
            obj = ModificationNetworkObject.from_securetrack(obj)
        else:
            _uid = None

        if not obj:
            return False

        obj_status = ObjectStatus(
            xsi_type=ObjectStatus.XsiType.EXISTING if _uid else ObjectStatus.XsiType.NEW
        )
        if _uid:
            obj_status.st_uid = _uid

        net_mod = NetworkObjectCellModification(
            action=cell_action,
            device_network_object=DeviceNetworkObject(
                device_id=device_id, status=obj_status
            ),
        )

        found = False
        for ref in objlist:
            if obj_status.st_uid and ref.uid == obj_status.st_uid:
                found = True
                break

        if (cell_action == CellAction.REMOVE and found) or (
            cell_action == CellAction.ADD and not found
        ):
            net_mod.device_network_object.network_object.append(obj)
            modlist.append(net_mod)
            return True
        else:
            return False

    def add_source(self, obj: Union[str, NetworkObject], device_id):
        if not self.source_modifications:
            self.source_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.source_modifications.network_object_cell_modifications,
            self.rule.src_networks,
            obj,
            device_id,
            CellAction.ADD,
        )

    def remove_source(self, obj: Union[str, NetworkObject], device_id):
        if not self.source_modifications:
            self.source_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.source_modifications.network_object_cell_modifications,
            self.rule.src_networks,
            obj,
            device_id,
            CellAction.REMOVE,
        )

    def add_destination(self, obj: Union[str, NetworkObject], device_id):
        if not self.destination_modifications:
            self.destination_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.destination_modifications.network_object_cell_modifications,
            self.rule.dest_networks,
            obj,
            device_id,
            CellAction.ADD,
        )

    def remove_destination(self, obj: Union[str, NetworkObject], device_id):
        if not self.destination_modifications:
            self.destination_modifications = NetworkObjectCellModifications()

        return self._add_net_obj_mod(
            self.destination_modifications.network_object_cell_modifications,
            self.rule.dest_networks,
            obj,
            device_id,
            CellAction.REMOVE,
        )

    def add_service(self, obj, device_id):
        if not self.service_modifications:
            self.service_modifications = ServiceObjectCellModifications()

        return self._add_service_obj_mod(
            self.service_modifications.service_object_cell_modifications,
            self.rule.src_services + self.rule.dest_services,
            obj,
            device_id,
            CellAction.ADD,
        )

    def remove_service(self, obj, device_id):
        if not self.service_modifications:
            self.service_modifications = ServiceObjectCellModifications()

        return self._add_service_obj_mod(
            self.service_modifications.service_object_cell_modifications,
            self.rule.src_services + self.rule.dest_services,
            obj,
            device_id,
            CellAction.REMOVE,
        )


def classify_rule_modification(obj):
    return (
        {"modify_rule_modification": ModifyRuleModification}
        .get(obj.get(Jsonable.Prop.XSI_TYPE.value))
        .kwargify(obj)
    )


@propify
class RuleModificationField(RuleOperation):
    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_MODIFICATION_FIELD, key=Jsonable.Prop.XSI_TYPE.value
    )

    rule_modifications: List[RuleModification] = prop(
        factory=list, flatify="rule_modification", kwargify=classify_rule_modification
    )

    def _get_rule_mod_for_rule(self, rule: SecurityRule):
        rule_key = RuleKey(
            device_id=rule.device.id,
            binding_uid=rule.bindings[0].uid,
            rule_uid=rule.uid,
        )

        for mod in self.rule_modifications:
            if mod.rule_key == rule_key:
                return mod

        mod = ModifyRuleModification(rule_key=rule_key)
        mod._rule = rule

        self.rule_modifications.append(mod)

        return mod

    def add_rule_modification(self, rule: SecurityRule) -> RuleModification:
        if not self.is_rule_added(rule):
            self.add_rule(rule)

        mod = self._get_rule_mod_for_rule(rule)
        return mod

    def add_source_object(self, rule: SecurityRule, obj: Union[str, NetworkObject]):
        return self.add_rule_modification(rule).add_source(obj, rule.device.id)

    def remove_source_object(self, rule: SecurityRule, obj: Union[str, NetworkObject]):
        return self.add_rule_modification(rule).remove_source(obj, rule.device.id)

    def add_destination_object(
        self, rule: SecurityRule, obj: Union[str, NetworkObject]
    ):
        return self.add_rule_modification(rule).add_destination(obj, rule.device.id)

    def remove_destination_object(
        self, rule: SecurityRule, obj: Union[str, NetworkObject]
    ):
        return self.add_rule_modification(rule).remove_destination(obj, rule.device.id)

    def add_service(self, rule: SecurityRule, obj: ModificationServiceObject):
        return self.add_rule_modification(rule).add_service(obj, rule.device.id)

    def remove_service(self, rule: SecurityRule, obj: ModificationServiceObject):
        return self.add_rule_modification(rule).remove_service(obj, rule.device.id)


@propify
class RuleRecertification(RuleOperation):
    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_RECERTIFICATION, key=Jsonable.Prop.XSI_TYPE.value
    )


@propify
class RuleDecommission(RuleOperation):
    class Action(Enum):
        DISABLE = "disable"
        REMOVE = "remove"

    xsi_type: FieldXsiType = prop(
        FieldXsiType.RULE_DECOMMISSION, key=Jsonable.Prop.XSI_TYPE.value
    )

    designer_result: Optional[Designer] = prop(None)
    verifier_result: Optional[dict] = prop(None)

    action: Optional[Action] = prop(None)



import pytest
import responses

from . import conftest  # noqa

from pytos2.secureapp.entrypoint import Sa
from pytos2.secureapp.application_identities import ApplicationIdentity


class TestEntrypoint:
    @responses.activate
    def test_application_identities(self, sa: Sa, application_identities_mock):
        # print(dir(sa))
        identity = sa.application_identities[0]
        assert isinstance(identity, ApplicationIdentity)

    @responses.activate
    def test_application_identites_error(self, sa: Sa, failure_mock):
        with pytest.raises(ValueError):
            sa.application_identities



from typing import Optional

from pytos2.utils import propify, prop
from pytos2.models import Jsonable


@propify
class TopologySyncStatus(Jsonable):
    percentage: Optional[int] = prop(None)
    description: Optional[str] = prop(None)



from typing import Optional, Union

from enum import Enum

import attr

from pytos2.utils import jsonify, kwargify, propify, prop


@propify
class Jsonable:
    class Prop(Enum):
        XSI_TYPE = "@xsi.type"

    data: Optional[dict] = attr.ib(None, eq=False, repr=False)
    id: Optional[int] = attr.ib(None, eq=False)
    xsi_type: Optional[Union[str, Enum]] = prop(
        None, key=Prop.XSI_TYPE.value, repr=False
    )
    _flatifies: dict = attr.ib(factory=dict, repr=False)

    _json_override: Optional[dict] = attr.ib(None, eq=False, repr=False, init=False)

    @property
    def _json(self) -> dict:
        return self._json_override or jsonify(self)

    @_json.setter
    def _json(self, val):
        self._json_override = val

    @classmethod
    def kwargify(cls, obj: dict):
        _obj, kwargs = kwargify(cls, obj)
        return cls(**kwargs)  # type: ignore

    def __getitem__(self, key):
        return getattr(self, key)


class UnMapped(dict):
    def __init__(self, obj: dict):
        super().__init__(obj)
        self._json = obj
        self.data = obj
        self._type = "UnMapped"

    @classmethod
    def kwargify(cls, obj: dict):
        return cls(obj)



import pytest
import json
import responses

from pytos2.securetrack.generic_ignored_interface import GenericIgnoredInterface
from pytos2.utils import get_api_node


class TestGenericIgnoredInterface:
    ignored_interfaces = json.load(
        open("tests/securetrack/json/generic_ignored_interfaces/mgmt-10.json")
    )
    interface = ignored_interfaces["IgnoredInterfaces"]

    @responses.activate
    def test_add_generic_ignored_interfaces_200(self, st):
        """Add one or multiple interface"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface",
            status=201,
        )

        """Add single interface"""
        newInterface = st.add_generic_ignored_interfaces(self.interface[0])
        assert newInterface == None

        """Add multiple interface"""
        newInterfaces = st.add_generic_ignored_interfaces(self.interface)
        assert newInterfaces == None

    @responses.activate
    def test_add_generic_ignored_interfaces_400(self, st):
        """Add one or multiple interface"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface",
            status=400,
        )

        """Add single interface"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_ignored_interfaces(self.interface[0])
        assert "Bad Request" in str(exception.value)

        """Add multiple interface"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_ignored_interfaces(self.interface)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_ignored_interfaces_404(self, st):
        """Add one or multiple interface"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface",
            status=404,
        )

        """Add single interface"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_ignored_interfaces(self.interface[0])
        assert "Not Found" in str(exception.value)

        """Add multiple interface"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_ignored_interfaces(self.interface)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_ignored_interfaces(self, st, generic_ignored_interface_mock):
        """Get interface by interface id"""
        interfaces = [
            GenericIgnoredInterface.kwargify(d)
            for d in get_api_node(
                self.ignored_interfaces, "IgnoredInterfaces", listify=True
            )
        ]
        interfaceByInt = st.get_generic_ignored_interfaces(10)
        interfaceByStr = st.get_generic_ignored_interfaces("10")
        assert interfaceByInt == interfaces
        assert interfaceByStr == interfaces

        with pytest.raises(ValueError) as exception:
            st.get_generic_ignored_interfaces(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_ignored_interfaces("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_ignored_interfaces(
        self, st, generic_ignored_interface_mock
    ):
        """Delete interface by interface id"""
        interfaceByInt = st.delete_generic_ignored_interfaces(10)
        interfaceByStr = st.delete_generic_ignored_interfaces("10")
        assert interfaceByInt == None
        assert interfaceByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_ignored_interfaces(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_ignored_interfaces("404")
        assert "Not Found" in str(exception.value)



from enum import Enum
from typing import Optional

from pytos2.api import get_app_api_session
from pytos2.utils import setup_logger


LOGGER = setup_logger("sa_api")


class SaAPI:
    class Meta(Enum):
        PATH = "securechangeworkflow/api/secureapp"
        APP = "SCW"
        TOS2_ENV = "SC_SERVER_SERVICE"

    def __init__(
        self, hostname: Optional[str], username: Optional[str], password: Optional[str]
    ):
        self.hostname, self.username, self.password, self.session = get_app_api_session(
            app=self, hostname=hostname, username=username, password=password
        )

    def get_application_identities(self):
        r = self.session.get("repository/application_identities")
        return r



import pytest
import json
import responses

from pytos2.securetrack.generic_vpn import GenericVpn
from pytos2.utils import get_api_node


class TestGenericVpn:
    device = json.load(open("tests/securetrack/json/generic_vpns/device-3.json"))
    vpns = device["GenericVpns"]

    vpn24 = json.load(open("tests/securetrack/json/generic_vpns/vpn-24.json"))
    vpn = vpn24["GenericVpn"]

    @responses.activate
    def test_add_generic_vpns_200(self, st):
        """Add one or multiple vpns"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=201,
        )

        """Add single vpn"""
        newVpn = st.add_generic_vpn(self.vpn)
        assert newVpn == None

        """Add multiple vpns"""
        newVpns = st.add_generic_vpns(self.vpns)
        assert newVpns == None

    @responses.activate
    def test_add_generic_vpns_400(self, st):
        """Add one or multiple vpns"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=400,
        )

        """Add single vpn"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_vpn(self.vpn)
        assert "Bad Request" in str(exception.value)

        """Add multiple vpns"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_vpns(self.vpns)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_vpns_404(self, st):
        """Add one or multiple vpns"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=404,
        )

        """Add single vpn"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_vpn(self.vpn)
        assert "Not Found" in str(exception.value)

        """Add multiple vpns"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_vpns(self.vpns)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_vpn(self, st, generic_vpn_mock):
        """Get vpn by vpn id"""
        mockRes = get_api_node(self.vpn24, "GenericVpn")
        vpn = GenericVpn.kwargify(mockRes)
        vpnByInt = st.get_generic_vpn(24)
        vpnByStr = st.get_generic_vpn("24")
        assert vpnByInt == vpn
        assert vpnByStr == vpn

        with pytest.raises(ValueError) as exception:
            st.get_generic_vpn(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_vpn("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_vpns(self, st, generic_vpn_mock):
        """Get vpns by device id"""
        vpns = [
            GenericVpn.kwargify(d)
            for d in get_api_node(self.device, "GenericVpns", listify=True)
        ]
        vpnsByInt = st.get_generic_vpns(3)
        vpnsByStr = st.get_generic_vpns("3")
        assert vpnsByInt == vpns
        assert vpnsByStr == vpns

        with pytest.raises(ValueError) as exception:
            st.get_generic_vpns(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_vpns("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_vpns_200(self, st):
        """Update one or multiple vpns"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=200,
        )

        """Update single vpn"""
        newVpn = st.update_generic_vpn(self.vpn)
        assert newVpn == None

        """Update multiple vpns"""
        newVpns = st.update_generic_vpns(self.vpns)
        assert newVpns == None

    @responses.activate
    def test_update_generic_vpns_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_vpn(self.vpn)
        assert "Bad Request" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_vpns(self.vpns)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_vpns_404(self, st):
        """PUT device id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_vpn(self.vpn)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_vpns(self.vpns)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_vpn(self, st, generic_vpn_mock):
        """Delete vpn by vpn id"""
        vpnByInt = st.delete_generic_vpn(24)
        vpnByStr = st.delete_generic_vpn("24")
        assert vpnByInt == None
        assert vpnByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_vpn(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_vpn("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_vpns(self, st, generic_vpn_mock):
        """Delete vpns by device id"""
        vpnsByInt = st.delete_generic_vpns(1)
        vpnsByStr = st.delete_generic_vpns("1")
        assert vpnsByInt == None
        assert vpnsByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_vpns(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_vpns("404")
        assert "Not Found" in str(exception.value)



from enum import Enum
from typing import Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop

from netaddr import IPAddress


@propify
class GenericIgnoredInterface(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int))
    interfaceName: str = prop(None)
    mgmtId: Optional[int] = prop(None, converter=optional(int))
    ip: Optional[IPAddress] = prop(None)



from typing import Union, Optional, Iterator, List
from enum import Enum

from traversify import Traverser
from pathlib import Path

# avoid circular imports
import pytos2
from .api import ScwAPI
from requests import Response
from requests.exceptions import HTTPError

from pytos2.securechange.user import classify_user_object, SCWParty, SCWUser
from pytos2.securechange.device import DeviceExclusions

from pytos2.utils import NoInstance, get_api_node
from pytos2.utils.cache import Cache, CacheIndex


class Scw:
    default: Union["Scw", NoInstance] = NoInstance(
        "Scw.default",
        "No Scw instance has been initialized yet, initialize with `Scw(*args, **kwargs)`",
    )

    def __init__(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        default=True,
    ):
        self.api: ScwAPI = ScwAPI(hostname, username, password)
        if default:
            Scw.default = self

        self.user_cache = Cache()
        self.users_by_name = self.user_cache.make_index("name")
        self.users_by_id = self.user_cache.make_index("id")

    def ticket_search(
        self,
        subject: Optional[str] = None,
        requester: Optional[str] = None,
        group: Optional[str] = None,
        assigned_to: Optional[str] = None,
        priority: Optional[str] = None,
        status: Optional[Union[str, "pytos2.securechange.ticket.Task.Status"]] = None,
        sla_status: Optional[
            Union[str, "pytos2.securechange.ticket.Ticket.SlaStatus"]
        ] = None,
        field_name: Optional[str] = None,
        field_value: Optional[str] = None,
        current_step: Optional[str] = None,
        expiration_date_from: Optional[str] = None,
        expiration_date_to: Optional[str] = None,
        domain_name: Optional[str] = None,
    ) -> List["pytos2.securechange.ticket.TicketSearchResult"]:
        # params are everything but self

        params = {key: value for (key, value) in locals().items() if key != "self"}

        from pytos2.securechange.ticket import (
            TicketSearchResult,
            TicketStatus,
            Ticket,
            Task,
        )

        for k, param in params.items():
            if isinstance(param, Enum):
                params[k] = param.value

        r = self.api.session.get("tickets/search", params=params)
        if not r.ok:
            r.raise_for_status()
        else:
            tickets = get_api_node(
                r.json(), "tickets_search_results.ticket_result", listify=True
            )
            return [TicketSearchResult.kwargify(t) for t in tickets]

    def get_users(
        self,
        show_indirect_relation: Optional[bool] = None,
        user_name: Optional[str] = None,
        email: Optional[str] = None,
        exact_name: Optional[bool] = None,
    ) -> List[SCWParty]:
        params = {}
        if show_indirect_relation:
            params["showIndirectRelation"] = show_indirect_relation
        if user_name:
            params["user_name"] = user_name
        if email:
            params["email"] = email
        if exact_name:
            params["exact_name"] = exact_name

        response = self.api.session.get("users", params=params)
        if not response.ok:
            response.raise_for_status()
        else:
            _json = response.json()
            users_node = get_api_node(_json, "users.user", listify=True)

            users = []
            self.user_cache.clear()

            for obj in users_node:
                user = classify_user_object(obj)
                users.append(user)

            self.user_cache.set_data(users)
            return users

    def _get_user_from_server(self, identifier: int) -> SCWParty:
        response = self.api.session.get(f"users/{identifier}")

        # if not response.ok:
        response.raise_for_status()
        # else:
        _json = response.json()

        key = ""
        if "group" in _json:
            key = "group"
        elif "user" in _json:
            key = "user"
        else:
            raise KeyError(
                f"Root user class key {_json.keys()} not currently supported by pytos2"
            )

        user_json = _json[key]
        if isinstance(user_json, list):
            user_json = user_json[0]
        return classify_user_object(user_json, obj_type=key)

    def get_user(
        self,
        identifier: Union[str, int],
        expand: bool = False,
        update_cache: Optional[bool] = None,
    ) -> SCWParty:
        if update_cache is not False and self.user_cache.is_empty():
            _ = self.get_users()  # create or update cache

        if isinstance(identifier, str):
            user = self.users_by_name.get(identifier)
            if not user:
                _ = self.get_users()
                user = self.users_by_name.get(identifier)
                if not user:
                    raise ValueError(f"User with name {identifier} not found")

            if expand:
                # this API only give @xsi.type as additional info
                return self._get_user_from_server(user.id)
            else:
                return user
        else:
            user = self.users_by_id.get(identifier)
            if not user:
                _ = self.get_users()
                user = self.users_by_id.get(identifier)
            if not user:
                raise ValueError(f"User with id {identifier} not found")

            if expand:
                try:
                    return self._get_user_from_server(identifier)
                except HTTPError as e:
                    # wrap the HTTPError into ValueError for consisency
                    raise ValueError(f"User with id {identifier} not found got {e}")
            else:
                return user

    def get_excluded_device_ids(self, show_all: Optional[bool] = None) -> List[int]:
        url = f"devices/excluded"
        if isinstance(show_all, bool):
            url += "?show_all="
            show_all = str(show_all).lower()
            url += show_all

        r = self.api.session.get(url)
        if r.ok:
            excludes_json = r.json()
            device_ids_model = DeviceExclusions.kwargify(excludes_json)
            return device_ids_model.device_ids
        else:
            r.raise_for_status()

    def get_tickets(
        self,
        status: Optional[
            Union[
                "pytos2.securechange.ticket.TicketStatus",
                List["pytos2.securechange.ticket.TicketStatus"],
            ]
        ] = None,
        start: Optional[int] = None,
        descending: Optional[bool] = None,
        expand_links: Optional[bool] = None,
    ) -> List["pytos2.securechange.ticket.TicketIterator"]:
        from pytos2.securechange.ticket import TicketStatus, TicketIterator

        params = {}

        if status is not None:
            if isinstance(status, list):
                params["status"] = ",".join([s.value for s in status])
            else:
                params["status"] = status.value
        if start is not None:
            params["start"] = start
        if descending is not None:
            params["desc"] = "true" if descending else "false"
        if expand_links is not None:
            params["expand_links"] = "true" if expand_links else "false"

        return list(TicketIterator(self.api.session, params))

    def get_ticket(self, _id: int) -> "pytos2.securechange.ticket.Ticket":
        from pytos2.securechange.ticket import Ticket

        r = self.api.session.get(f"tickets/{_id}")
        if r.ok:
            tkt = Ticket.kwargify(r.json())
            return tkt
        else:
            r.raise_for_status()

    def reassign_ticket(
        self,
        ticket,
        user,
        step: Union[None, "Step", int, str] = None,
        task: Union[None, "Task", int] = None,
        comment="",
    ) -> None:
        from pytos2.securechange.ticket import Step, Task, Ticket

        if not isinstance(ticket, Ticket):
            ticket = self.get_ticket(ticket)

        if not isinstance(user, SCWUser):
            user = Scw.default.get_user(user)

        if step is None:
            step = ticket.current_step

        if not isinstance(step, Step):
            step = ticket.get_step(step)

        if task is None:
            task = step.get_task(0)

        if not isinstance(task, Task):
            task = step.get_task(task)

        try:
            response = self.api.session.put(
                f"tickets/{ticket.id}/steps/{step.id}/tasks/{task.id}/reassign/{user.id}",
                json={"reassign_task_comment": {"comment": comment}},
            )
            if not response.ok:
                msg = response.json().get("result").get("message")
                response.raise_for_status()
        except HTTPError as e:
            raise ValueError(
                f"Got {e}, with Error Message: {msg}. Only tasks under current step can be reassigned"
            )

    def get_attachment(self, file_id: Union[int, str]) -> bytes:
        response = self.api.session.get(f"attachments/{file_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Attachment: {e}")
        return response.content

    def add_attachment(self, file: str) -> str:
        attachment = {"attachment": open(file, "rb")}
        response = self.api.session.post("attachments", files=attachment)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Attachment: {e}")
        return response.text

    def add_comment(
        self,
        ticket_id: Union[int, str],
        step_id: Union[int, str],
        task_id: Union[int, str],
        comment_content: str,
        attachment_uuids: Optional[List[str]],
    ) -> str:
        def _format_attachments(uuids: List[str]) -> dict:
            return {"attachment": [{"uid": id} for id in uuids]}

        comment = {"comment": {"content": comment_content}}
        if attachment_uuids:
            comment["comment"]["attachments"] = _format_attachments(attachment_uuids)

        response = self.api.session.post(
            f"tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/comments",
            json=comment,
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Comment: {e}")

        return response.text

    def delete_comment(
        self,
        ticket_id: Union[int, str],
        comment_id: Union[int, str],
    ) -> None:
        response = self.api.session.delete(
            f"tickets/{ticket_id}/comments/{comment_id}",
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Comment: {e}")



from enum import Enum
from typing import List, Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop


@propify
class GenericTransparentFirewall(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int))
    outputL3DeviceId: Optional[int] = prop(None, converter=optional(int))
    outputL3IsGenericDevice: Optional[bool] = prop(None)
    outputL3InterfaceName: Optional[str] = prop(None)
    layer2DeviceId: Optional[int] = prop(None, converter=optional(int))
    inputL2InterfaceName: Optional[str] = prop(None)
    outputL2InterfaceName: Optional[str] = prop(None)
    inputL3DeviceId: Optional[int] = prop(None, converter=optional(int))
    inputL3IsGenericDevice: Optional[bool] = prop(None)
    inputL3InterfaceName: Optional[str] = prop(None)



from enum import Enum
from typing import Optional, List

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop

from netaddr import IPAddress


@propify
class JoinCloud(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int))
    name: str = prop(None)
    clouds: List[int] = prop(factory=list)



import pytest
import json
import responses

from pytos2.securetrack.generic_interface_customer import GenericInterfaceCustomer
from pytos2.utils import get_api_node


class TestGenericInterfaceCustomer:
    device = json.load(
        open("tests/securetrack/json/generic_interface_customers/device-5.json")
    )
    interfaces = device["InterfaceCustomerTags"]

    interface_customer = json.load(
        open("tests/securetrack/json/generic_interface_customers/int-cust-74.json")
    )
    interface = interface_customer["InterfaceCustomerTag"]

    @responses.activate
    def test_add_generic_interface_customers_200(self, st):
        """Add one or multiple interface customers"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=201,
        )

        """Add single interface customer"""
        newRoute = st.add_generic_interface_customer(self.interface)
        assert newRoute == None

        """Add multiple interface customers"""
        newRoutes = st.add_generic_interface_customers(self.interfaces)
        assert newRoutes == None

    @responses.activate
    def test_add_generic_interface_customers_400(self, st):
        """Add one or multiple interface customers"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=400,
        )

        """Add single interface customer"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_interface_customer(self.interface)
        assert "Bad Request" in str(exception.value)

        """Add multiple interface customers"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_interface_customers(self.interfaces)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_interface_customers_404(self, st):
        """Add one or multiple interface customers"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=404,
        )

        """Add single interface customer"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_interface_customer(self.interface)
        assert "Not Found" in str(exception.value)

        """Add multiple interface customers"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_interface_customers(self.interfaces)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_interface_customer(self, st, generic_interface_customer_mock):
        """Get interface by interface id"""
        mockRes = get_api_node(self.interface_customer, "InterfaceCustomerTag")
        interface = GenericInterfaceCustomer.kwargify(mockRes)
        interface_customerByInt = st.get_generic_interface_customer(74)
        interface_customerByStr = st.get_generic_interface_customer("74")
        assert interface_customerByInt == interface
        assert interface_customerByStr == interface

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface_customer(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface_customer("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_interface_customers(self, st, generic_interface_customer_mock):
        """Get interfaces by device id"""
        interfaces = [
            GenericInterfaceCustomer.kwargify(d)
            for d in get_api_node(self.device, "InterfaceCustomerTags", listify=True)
        ]
        interface_customersByInt = st.get_generic_interface_customers(5)
        interface_customersByStr = st.get_generic_interface_customers("5")
        assert interface_customersByInt == interfaces
        assert interface_customersByStr == interfaces

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface_customers(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface_customers("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_interface_customers_200(self, st):
        """Update one or multiple interface customers"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=200,
        )

        """Update single interface customer"""
        newRoute = st.update_generic_interface_customer(self.interface)
        assert newRoute == None

        """Update multiple interface customers"""
        newRoutes = st.update_generic_interface_customers(self.interfaces)
        assert newRoutes == None

    @responses.activate
    def test_update_generic_interface_customers_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_interface_customer(self.interface)
        assert "Bad Request" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_interface_customers(self.interfaces)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_interface_customers_404(self, st):
        """PUT device id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_interface_customer(self.interface)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_interface_customers(self.interfaces)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_interface_customer(
        self, st, generic_interface_customer_mock
    ):
        """Delete interface by interface id"""
        interface_customerByInt = st.delete_generic_interface_customer(74)
        interface_customerByStr = st.delete_generic_interface_customer("74")
        assert interface_customerByInt == None
        assert interface_customerByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_interface_customer(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_interface_customer("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_interface_customers(
        self, st, generic_interface_customer_mock
    ):
        """Delete interfaces by device id"""
        interface_customersByInt = st.delete_generic_interface_customers(5)
        interface_customersByStr = st.delete_generic_interface_customers("5")
        assert interface_customersByInt == None
        assert interface_customersByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_interface_customers(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_interface_customers("404")
        assert "Not Found" in str(exception.value)



import pytest
import json
import responses

from pytos2.securetrack.generic_transparent_firewall import GenericTransparentFirewall
from pytos2.utils import get_api_node


class TestGenericTransparentFirewall:
    device = json.load(
        open("tests/securetrack/json/generic_transparent_firewalls/device-9.json")
    )
    firewalls = device["TransparentFirewalls"]

    @responses.activate
    def test_add_generic_transparent_firewalls_200(self, st):
        """Add one or multiple firewalls"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw",
            status=201,
        )

        """Add single firewall"""
        newFirewall = st.add_generic_transparent_firewalls(self.firewalls[0])
        assert newFirewall == None

        """Add multiple firewalls"""
        newFirewalls = st.add_generic_transparent_firewalls(self.firewalls)
        assert newFirewalls == None

    @responses.activate
    def test_add_generic_transparent_firewalls_400(self, st):
        """Add one or multiple firewalls"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw",
            status=400,
        )

        """Add single firewall"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_transparent_firewalls(self.firewalls[0])
        assert "Bad Request" in str(exception.value)

        """Add multiple firewalls"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_transparent_firewalls(self.firewalls)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_transparent_firewalls_404(self, st):
        """Add one or multiple firewalls"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw",
            status=404,
        )

        """Add single firewall"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_transparent_firewalls(self.firewalls[0])
        assert "Not Found" in str(exception.value)

        """Add multiple firewalls"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_transparent_firewalls(self.firewalls)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_transparent_firewalls(
        self, st, generic_transparent_firewall_mock
    ):
        """Get firewalls by device id"""
        firewalls = [
            GenericTransparentFirewall.kwargify(d)
            for d in get_api_node(self.device, "TransparentFirewalls", listify=True)
        ]
        firewallsByInt = st.get_generic_transparent_firewalls(9)
        firewallsByStr = st.get_generic_transparent_firewalls("9")
        assert firewallsByInt == firewalls
        assert firewallsByStr == firewalls

        with pytest.raises(ValueError) as exception:
            st.get_generic_transparent_firewalls(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_transparent_firewalls("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_transparent_firewalls_200(self, st):
        """Update one or multiple firewalls"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw",
            status=200,
        )

        """Update single firewall"""
        newFirewall = st.update_generic_transparent_firewalls(self.firewalls[0])
        assert newFirewall == None

        """Update multiple firewalls"""
        newFirewalls = st.update_generic_transparent_firewalls(self.firewalls)
        assert newFirewalls == None

    @responses.activate
    def test_update_generic_transparent_firewalls_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_transparent_firewalls(self.firewalls[0])
        assert "Bad Request" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_transparent_firewalls(self.firewalls)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_transparent_firewalls_404(self, st):
        """PUT device id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_transparent_firewalls(self.firewalls[0])
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_transparent_firewalls(self.firewalls)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_transparent_firewall(
        self, st, generic_transparent_firewall_mock
    ):
        """Delete firewall by firewall id"""
        firewallByInt = st.delete_generic_transparent_firewall(23)
        firewallByStr = st.delete_generic_transparent_firewall("23")
        assert firewallByInt == None
        assert firewallByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_transparent_firewall(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_transparent_firewall("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_transparent_firewalls(
        self, st, generic_transparent_firewall_mock
    ):
        """Delete firewalls by device id"""
        firewallsByInt = st.delete_generic_transparent_firewalls(9)
        firewallsByStr = st.delete_generic_transparent_firewalls("9")
        assert firewallsByInt == None
        assert firewallsByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_transparent_firewalls(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_transparent_firewalls("404")
        assert "Not Found" in str(exception.value)



from collections import OrderedDict
from datetime import date
from typing import Dict, Iterable, Iterator, List, Tuple, Optional, Union
from io import BytesIO
import typing

from requests.exceptions import HTTPError
from requests import Response

from pytos2.securetrack.generic_transparent_firewall import GenericTransparentFirewall

from .api import StAPI
from .device import Device
from .domain import Domain
from .network_object import classify_network_object, NetworkObject
from .policy_browser import Emptiness
from .revision import Revision
from .rule import BindingPolicy, Documentation, SecurityRule
from pytos2.utils import (
    NoInstance,
    get_api_node,
    sanitize_uid,
    uids_match,
    setup_logger,
)
from pytos2.utils.cache import Cache
from .service_object import classify_service_object, Service
from .zone import Zone, ZoneReference, ZoneEntry
from .interface import Interface, BindableObject, TopologyInterface
from .generic_device import GenericDevice
from .generic_ignored_interface import GenericIgnoredInterface
from .generic_interface_customer import GenericInterfaceCustomer
from .generic_interface import GenericInterface
from .generic_route import GenericRoute
from .generic_vpn import GenericVpn
from .join_cloud import JoinCloud
from .topology import TopologySyncStatus

LOGGER = setup_logger("st_entrypoint")


def _bool(x: bool) -> str:
    return "true" if x else "false"


def _querify(k: str, v: Union[str, bool, List[Union[str, bool]]]) -> str:
    # Used in `St.search_rules`. See that method for more details.

    # `search_text` params (key:values pairs with semantic key meanings to
    # Policy Browser, such as `'action:accept'`) are specified in the URI in
    # the format: `'key:value+key:value+...'`.  (e.g.:
    # `uid:123+action:accept`), so we have to marshal the given params into
    # said format.
    #
    # `strs remains strings, `bool`s are converted to the string `"true"` or
    # `"false"` respectively, and array values are converted to look like:
    # `'key:value1+key:value2+...'`.
    if isinstance(v, list):
        return " ".join([_querify(k, v_) for v_ in v])
    elif isinstance(v, bool):
        return f"{k}:{_bool(v)}"
    else:
        return f"{k}:{v}"


class St:
    default: Union["St", NoInstance] = NoInstance(
        "St.default",
        "No St instance has been initialized yet, initialize with `St(*args, **kwargs)`",
    )

    def __init__(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        default=True,
        cache=True,
    ):
        self.api: StAPI = StAPI(hostname, username, password)
        if default:
            St.default = self
        self.cache = cache

        self._devices_cache = Cache()
        self._devices_index = self._devices_cache.make_index(["name", "id"])

        self._network_objects_by_device_id_by_name: dict = {}
        self._network_objects_by_uid: dict = {}
        self._services_by_uid: dict = {}
        self._services_by_device_id_by_name: dict = {}
        self._device_rules_dict: dict = {}
        self._revision_rules_dict: dict = {}
        self._revisions_dict: dict = {}
        self._device_revisions_dict: dict = {}
        self._rules_dict: dict = {}
        self._zones_cache = Cache()
        self._zones_index = self._zones_cache.make_index(["name", "id"])
        self._zones: list = []
        self._zones_dict: dict = {}
        self._domains_cache = Cache()
        self._domains_index = self._domains_cache.make_index(["name", "id"])
        self._domains: list = []
        self._domains_dict: dict = {}

        self._generic_devices_cache = Cache()
        self._generic_devices_index = self._generic_devices_cache.make_index(
            ["name", "id"]
        )

    def _prime_generic_devices_cache(self):
        generic_devices = self.get_generic_devices(cache=False)

        self._generic_devices_cache.clear()
        for d in generic_devices:
            self._generic_devices_cache.add(d)

    def _prime_domains_cache(self):
        domains = self.api.session.get("domains").json()

        self._domains_cache.clear()
        for domain in get_api_node(domains, "domain", listify=True):
            domain_obj = Domain.kwargify(domain)
            self._domains_cache.add(domain_obj)

    def _prime_zones_cache(self):
        zones = self.api.session.get("zones").json()

        self._zones_cache.clear()
        for zone in get_api_node(zones, "zones.zone", listify=True):
            zone_obj = Zone.kwargify(zone)
            self._zones_cache.add(zone_obj)

    def _prime_devices_cache(self):
        devices = self.api.session.get("devices").json()

        self._devices_cache.clear()
        for device in get_api_node(devices, "devices.device", listify=True):
            device_obj = Device.kwargify(device)
            self._devices_cache.add(device_obj)

    def get_devices(self, cache: Optional[bool] = None, filter: Optional[dict] = None):
        if cache is not False and self.cache:
            if self._devices_cache.is_empty():
                self._prime_devices_cache()

            if not filter:
                return self._devices_cache.get_data()
            else:
                return [
                    d
                    for d in self._devices_cache.get_data()
                    if set(filter.items()).issubset(set(d.data.items()))
                ]
        else:
            devices = self.api.session.get("devices").json()
            d_list = [
                Device.kwargify(d)
                for d in get_api_node(devices, "devices.device", listify=True)
            ]
            if not filter:
                return d_list
            else:
                return [
                    d
                    for d in d_list
                    if set(filter.items()).issubset(set(d.data.items()))
                ]

    def get_zones(self, cache: Optional[bool] = None):
        if cache is not False and self.cache:
            if self._zones_cache.is_empty():
                self._prime_zones_cache()
            return self._zones_cache.get_data()
        else:
            zones = self.api.session.get("zones").json()
            return [
                Zone.kwargify(d)
                for d in get_api_node(zones, "zones.zone", listify=True)
            ]

    def _resolve_zone_from_name(self, name: Union[str, List[str]]):
        zones = self.get_zones()
        if isinstance(name, str):
            name = [name]

        objects = []

        for n in name:
            for zone in zones:
                if zone.name == n:
                    objects.append(zone)

        return objects

    def get_zone_subnets(
        self, identifier: Union[int, str, List[int]]
    ) -> List[ZoneEntry]:
        def _send_request(id_list):
            _identifier = ",".join([str(i) for i in id_list])
            response = self.api.session.get(f"zones/{_identifier}/entries")
            if not response.ok:
                try:
                    msg = response.json().get("result").get("message")
                    response.raise_for_status()
                except HTTPError as e:
                    raise ValueError(
                        f"wrong zone identifier, got '{msg}' from API Error: {e}"
                    )
            else:
                zone_entries = get_api_node(
                    response.json(), "zone_entries.zone_entry", listify=True
                )
                zone_subnets = []
                for entry in zone_entries:
                    zone_subnets.append(ZoneEntry.kwargify(entry))

                return zone_subnets

        def _get(_identifier):
            if isinstance(_identifier, str):
                _identifier = [z.id for z in self._resolve_zone_from_name(_identifier)]

            if isinstance(_identifier, (list, int)):
                if isinstance(_identifier, int):
                    _identifier = [_identifier]

                if isinstance(_identifier, list):
                    i = 0
                    length = len(_identifier)
                    res_subnets = []
                    while i < length:
                        id_list = _identifier[i : i + 10]
                        i += 10

                        for entry in _send_request(id_list):
                            res_subnets.append(entry)
                            # yield entry

                    return res_subnets  # noqa
            else:
                raise TypeError(
                    f"input identifier can only be list, int, str or list[int]] but got {_identifier}"
                )

        zone_subnets = []
        for entry in _get(identifier):
            zone_subnets.append(entry)
        return zone_subnets

    def get_zone_descendants(
        self, identifier: Union[int, str, List[int]]
    ) -> List[ZoneReference]:
        def _send_request(id_list):
            _identifier = ",".join([str(i) for i in id_list])

            response = self.api.session.get(f"zones/{_identifier}/descendants")
            if not response.ok:
                try:
                    msg = response.json().get("result").get("message")
                    response.raise_for_status()
                except HTTPError as e:
                    raise ValueError(
                        f"wrong zone identifier, got '{msg}' from API Error: {e}"
                    )
            else:
                zones = get_api_node(response.json(), "zones.zone", listify=True)
                if not zones:
                    raise ValueError(f"can not find zones by given ids: {_identifier}")
                zone_ref = []
                for zone in zones:
                    zone_ref.append(ZoneReference.kwargify(zone))

                return zone_ref

        def _get(_identifier):
            if isinstance(_identifier, str):
                _identifier = [z.id for z in self._resolve_zone_from_name(_identifier)]

            if isinstance(_identifier, (list, int)):
                if isinstance(_identifier, int):
                    _identifier = [_identifier]

                if isinstance(_identifier, list):
                    i = 0
                    length = len(_identifier)
                    res_descendants = []
                    while i < length:
                        id_list = _identifier[i : i + 10]
                        i += 10

                        for zone in _send_request(id_list):
                            res_descendants.append(zone)

                    return res_descendants
            else:
                raise TypeError(
                    f"input identifier can only be list, int, str or list[int]] but got {_identifier}"
                )

        zone_descendants = []
        for zone in _get(identifier):
            zone_descendants.append(zone)
        return zone_descendants

    def get_zone(
        self, identifier: Union[int, str], cache: Optional[bool] = None
    ) -> Optional[Zone]:
        def _get(_identifier):
            if isinstance(_identifier, int):
                zone = get_api_node(self.api.get_zone_by_id(_identifier).json(), "zone")
                return Zone.kwargify(zone) if zone else None
            else:
                for zone in get_api_node(
                    self.api.get_zones_by_name(_identifier).json(),
                    "zones.zone",
                    default=[],
                ):
                    zone_obj = Zone.kwargify(zone)
                    if zone_obj.name == _identifier:
                        return zone_obj

        if cache is not False and self.cache:
            if self._zones_cache.is_empty():
                self._prime_zones_cache()
            return self._zones_index.get(identifier)
        else:
            return _get(identifier)

    def get_domains(self, cache: Optional[bool] = None):
        if cache is not False and self.cache:
            if self._domains_cache.is_empty():
                self._prime_domains_cache()
            return self._domains_cache.get_data()
        else:
            domains = self.api.session.get("domains").json()

            return [
                Domain.kwargify(d)
                for d in get_api_node(domains, "domain", listify=True)
            ]

    def get_domain(
        self, identifier: Union[int, str], cache: Optional[bool] = None
    ) -> Optional[Domain]:
        def _get(_identifier):
            if isinstance(_identifier, int):
                domain = get_api_node(
                    self.api.get_domain_by_id(_identifier).json(), "domain"
                )
                return Domain.kwargify(domain) if domain else None
            else:
                for domain in get_api_node(
                    self.api.get_domains_by_name(_identifier).json(),
                    "domain",
                    default=[],
                ):
                    domain_obj = Domain.kwargify(domain)
                    if domain_obj.name == _identifier:
                        return domain_obj

        if cache is not False and self.cache:
            if self._domains_cache.is_empty():
                self._prime_domains_cache()
            return self._domains_index.get(identifier)
        else:
            return _get(identifier)

    def add_domain(
        self,
        name: str,
        description: Optional[str] = None,
        address: Optional[str] = None,
    ) -> Optional[Domain]:
        res = self.api.post_domain(name=name, description=description, address=address)
        if res.ok:
            created_url = res.headers.get("Location", "")
            did = int(created_url.split("/")[-1])
            new_domain = get_api_node(self.api.get_domain_by_id(did).json(), "domain")
            if new_domain:
                return Domain.kwargify(new_domain)
            else:
                raise ValueError(
                    f"domain id: {did} not found by GET call after POSTing to SecureTrack"
                )

        else:  # pragma: no cover
            try:
                msg = res.json().get("result").get("message")
                res.raise_for_status()
            except HTTPError as e:
                raise ValueError(
                    f"unable to POST new domain :{name} to SecureTrack, got {msg} from API Error: {e}"
                )

    def update_domain(
        self,
        identifier: Union[int, str],
        name: Optional[str] = None,
        description: Optional[str] = None,
        address: Optional[str] = None,
    ) -> Optional[Domain]:
        if self._domains_cache.is_empty():
            self._prime_domains_cache()
        modify_domain = self._domains_index.get(identifier)
        res = self.api.put_domain(
            id=modify_domain.id,
            name=name or modify_domain.name,
            description=description or modify_domain.description,
            address=address or modify_domain.address,
        )
        if res.ok:
            modified_domain_json = get_api_node(
                self.api.get_domain_by_id(modify_domain.id).json(), "domain"
            )
            modified_domain = Domain.kwargify(modified_domain_json)
            self._domains_dict[identifier] = modified_domain
            return modified_domain

        else:
            try:
                msg = res.json().get("result").get("message")
                res.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Got {e}, with Error Message: {msg}")

    def get_device(
        self, identifier: Union[int, str], cache: Optional[bool] = None
    ) -> Optional[Device]:
        def _get(_identifier):
            if isinstance(_identifier, int):
                device = get_api_node(
                    self.api.get_device_by_id(identifier).json(), "device"
                )
                return Device.kwargify(device) if device else None
            else:
                for device in get_api_node(
                    self.api.get_devices_by_name(_identifier).json(),
                    "devices.device",
                    default=[],
                ):
                    device_obj = Device.kwargify(device)
                    if device_obj.name == _identifier:
                        return device_obj

        if cache is not False and self.cache:
            if self._devices_cache.is_empty():
                self._prime_devices_cache()
            return self._devices_index.get(identifier)
        else:
            return _get(identifier)

    def _prime_network_objects_cache(self, device_id: int):
        # bust existing cache for device_id
        self._network_objects_by_device_id_by_name[device_id] = {}
        self._network_objects_by_uid = {
            u: [o for o in objs if o.device_id != device_id]
            for u, objs in self._network_objects_by_uid.items()
        }

        network_objects = self.api.session.get(
            f"devices/{device_id}/network_objects"
        ).json()
        for obj in get_api_node(
            network_objects, "network_objects.network_object", listify=True
        ):
            obj = classify_network_object(dict(**obj, device_id=device_id))
            self._network_objects_by_device_id_by_name.setdefault(device_id, {})[
                str(obj["name"])
            ] = obj
            """
            self._network_objects_by_device_id_by_name.setdefault(device_id, {})[
                str(obj["display_name"])
            ] = obj
            """
            self._network_objects_by_uid.setdefault(
                sanitize_uid(obj["uid"]), []
            ).append(obj)

    def _prime_services_cache(self, device_id: int):
        # bust existing cache for device_id
        self._services_by_device_id_by_name[device_id] = {}
        self._services_by_uid = {
            u: [o for o in objs if o.device_id != device_id]
            for u, objs in self._services_by_uid.items()
        }

        services = self.api.session.get(f"devices/{device_id}/services").json()
        for obj in get_api_node(services, "services.service", listify=True):
            obj = classify_service_object(dict(**obj, device_id=device_id))
            self._services_by_device_id_by_name.setdefault(device_id, {})[
                str(obj["name"])
            ] = obj
            self._services_by_uid.setdefault(sanitize_uid(obj["uid"]), []).append(obj)

    def get_shadowing_rules_for_device(
        self: "St", device: str, rules: Iterator[str]
    ) -> List[Tuple[SecurityRule, List[SecurityRule]]]:
        rules = self.api.session.get(
            f"devices/{device}/shadowing_rules",
            params={"shadowed_uids": ",".join(rules)},
        )

        if not rules.ok:
            try:
                msg = rules.json().get("result").get("message")
                rules.raise_for_status()
            except HTTPError as e:
                raise ValueError(
                    f"Unable to get shadowing rules got '{msg}' from API Error: {e}"
                )
        rules_json = (
            rules.json()
            .get("cleanup_set")
            .get("shadowed_rules_cleanup")
            .get("shadowed_rules")
            .get("shadowed_rule")
        )
        result_rules = []
        for rule in rules_json:
            rule_and_shadowing_rules_pair = (
                SecurityRule.kwargify(rule.get("rule")),
                [
                    SecurityRule.kwargify(r)
                    for r in rule.get("shadowing_rules").get("rule")
                ],
            )
            result_rules.append(rule_and_shadowing_rules_pair)
        return result_rules

    def get_network_objects(
        self, device: Union[int, str], cache: Optional[bool] = None
    ):
        if cache is not False and self.cache:
            device_obj = self.get_device(device)
            if device_obj is None:
                raise ValueError(f"Device {device} not found")
            device_id = device_obj.id
            if device_id not in self._network_objects_by_device_id_by_name:
                self._prime_network_objects_cache(device_id)

            objs = list(self._network_objects_by_device_id_by_name[device_id].values())
            for obj in objs:
                obj.device_id = device_id
        return objs

    def get_network_object(
        self,
        name: Optional[str] = None,
        device: Union[int, str, None] = None,
        uid: Optional[str] = None,
        cache=True,
    ) -> NetworkObject:
        device_obj = None
        if device:
            device_obj = self.get_device(device)
            if device_obj is None:
                raise ValueError(f"Device {device} not found")
        if self.cache and cache:
            if uid:
                uid = sanitize_uid(uid)
                if uid not in self._network_objects_by_uid:
                    objs = self.api.session.get(
                        "network_objects/search", params={"filter": "uid", "uid": uid}
                    ).json()

                    for obj_json in get_api_node(
                        objs, "network_objects.network_object", listify=True
                    ):
                        obj = classify_network_object(obj_json)
                        self._network_objects_by_uid.setdefault(
                            sanitize_uid(obj["uid"]), []
                        ).append(obj)
                        self._network_objects_by_device_id_by_name.setdefault(
                            obj["device_id"], {}
                        )[obj["name"]] = obj

                objs = self._network_objects_by_uid.get(uid, [None])
                if len(objs) > 1:
                    if device_obj is None:
                        raise AssertionError(
                            f"More than one object found for uid {uid}, device argument must be passed"
                        )
                    else:
                        for obj in objs:
                            if obj.device_id == device_obj.id:
                                objs = [obj]
                                break
                        else:
                            objs = [None]
                return objs[0]

            elif not name or device_obj is None:
                raise ValueError(
                    "name and device arguments must be passed if uid is None"
                )
            device_id = device_obj.id
            if device_id not in self._network_objects_by_device_id_by_name:
                network_objects = self.api.session.get(
                    f"devices/{device_id}/network_objects"
                ).json()
                for obj_json in get_api_node(
                    network_objects, "network_objects.network_object", listify=True
                ):
                    obj = classify_network_object(obj_json)
                    self._network_objects_by_device_id_by_name.setdefault(
                        device_id, {}
                    )[str(obj["name"])] = obj
                    obj.device_id = device_id
                    self._network_objects_by_uid.setdefault(
                        sanitize_uid(obj["uid"]), []
                    ).append(obj)
            return self._network_objects_by_device_id_by_name[device_id].get(name)
        else:
            raise NotImplementedError(
                "Non-caching mode is not supported...yet"
            )  # pragma: no cover

    def get_services(self, device: Union[int, str], cache: Optional[bool] = None):
        if cache is not False and self.cache:
            device_obj = self.get_device(device)
            if device_obj is None:
                raise ValueError(f"Device {device} not found")
            device_id = device_obj.id
            if device_id not in self._services_by_device_id_by_name:
                self._prime_services_cache(device_id)
            objs = list(self._services_by_device_id_by_name[device_id].values())
            for obj in objs:
                obj.device_id = device_id
        return objs

    def get_service(
        self,
        name: Optional[str] = None,
        device: Union[int, str, None] = None,
        uid: Optional[str] = None,
        cache=True,
    ) -> Service:
        device_obj = None
        if device:
            device_obj = self.get_device(device)
            if device_obj is None:
                raise ValueError(f"Device {device} not found")
        if self.cache and cache:
            if uid:
                uid = sanitize_uid(uid)
                if uid not in self._services_by_uid:
                    params = {"filter": "uid", "uid": uid}

                    if device_obj:
                        params["device_id"] = device_obj.id

                    objs = self.api.session.get("services/search", params=params).json()

                    for obj_json in get_api_node(
                        objs, "services.service", listify=True
                    ):
                        obj = classify_service_object(obj_json)
                        self._services_by_uid.setdefault(
                            sanitize_uid(obj["uid"]), []
                        ).append(obj)

                        self._services_by_device_id_by_name.setdefault(
                            obj.device_id or device_obj.id, {}
                        )[obj["name"]] = obj

                objs = self._services_by_uid.get(uid, [None])
                if len(objs) > 1:
                    if device_obj is None:
                        raise AssertionError(
                            "More than one object found for uid {uid}, device argument must be passed"
                        )
                    else:
                        for obj in objs:
                            if obj.device_id == device_obj.id:
                                objs = [obj]
                                break
                        else:
                            objs = [None]
                return objs[0]

            elif not name or device_obj is None:
                raise ValueError(
                    "name and device arguments must be passed if uid is None"
                )
            device_id = device_obj.id
            if (
                device_id not in self._services_by_device_id_by_name
                or name not in self._services_by_device_id_by_name[device_id]
            ):
                services = self.api.session.get(f"devices/{device_id}/services").json()
                for obj_json in get_api_node(
                    services, "services.service", listify=True
                ):
                    obj = classify_service_object(obj_json)
                    self._services_by_device_id_by_name.setdefault(device_id, {})[
                        str(obj["name"])
                    ] = obj
                    obj.device_id = device_id
                    self._services_by_uid.setdefault(
                        sanitize_uid(obj["uid"]), []
                    ).append(obj)

            return self._services_by_device_id_by_name[device_id].get(name)
        else:
            raise NotImplementedError(
                "Non-caching mode is not supported...yet"
            )  # pragma: no cover

    def _prime_rules_cache(self):
        self._device_rules_dict = {}
        self._revision_rules_dict = {}

    def _transform_rules_response(self, rules_response: Response) -> Iterator:
        if not rules_response.ok:
            try:
                msg = rules_response.text
                rules_response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Got '{msg}' from Error :{e}")

        rules_json = rules_response.json()
        rules_json = get_api_node(rules_json, "rules.rule", listify=True)
        rules = [SecurityRule.kwargify(rule) for rule in rules_json]
        return rules

    def _transform_nat_rules_response(self, nat_rules_response: Response) -> Iterator:
        if not nat_rules_response.ok:
            try:
                msg = nat_rules_response.json().get("result").get("message")
                nat_rules_response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Got '{msg}' from Error: {e}")

        nat_rules_json = nat_rules_response.json()
        nat_rules_json = get_api_node(nat_rules_json, "nat_rule", listify=True)
        nat_rules = [SecurityRule.kwargify(nat_rule) for nat_rule in nat_rules_json]
        return nat_rules

    def _filter_rule_uid(self, rules, rule_uid):
        if rule_uid:
            mat_rules = [rule for rule in rules if uids_match(rule.uid, rule_uid)]
            return mat_rules
        else:
            return rules

    def _get_rules_by_revision(
        self,
        revision: int,
        rule_uid: Optional[str] = None,
        uid: Optional[str] = None,
        documentation: bool = True,
        cache: bool = True,
    ):
        revision_obj = self.get_revision(revision=revision)
        revision_id = revision_obj.id
        if cache and self.cache:
            if not self._revision_rules_dict:
                self._prime_rules_cache()

            if revision_id in self._revision_rules_dict:
                rules = self._revision_rules_dict[revision_id]

                if rule_uid is not None:
                    return self._filter_rule_uid(rules, rule_uid)
                else:
                    return rules
            else:
                rules_response = self.api.get_rules_from_revision_id(
                    revision_id, uid=uid, documentation=documentation
                )
                rules = self._transform_rules_response(rules_response)

                self._revision_rules_dict[revision_id] = rules
                return self._filter_rule_uid(rules, rule_uid)
        else:
            rules_response = self.api.get_rules_from_revision_id(
                revision_id, uid=uid, documentation=documentation
            )
            rules = self._transform_rules_response(rules_response)

            return self._filter_rule_uid(rules, rule_uid)

    def _get_rules_by_device(
        self,
        device: Union[str, int],
        rule_uid: Optional[str] = None,
        uid: Optional[str] = None,
        documentation: bool = True,
        cache: bool = True,
    ):
        device_obj = self.get_device(device)

        if device_obj is None:
            raise ValueError(f"Device {device} not found")

        latest_revision_id = device_obj.latest_revision
        device_id = device_obj.id
        if cache and self.cache:
            if not self._device_rules_dict:
                self._prime_rules_cache()

            if device_id in self._device_rules_dict:
                return self._filter_rule_uid(
                    self._device_rules_dict[device_id], rule_uid
                )

            else:
                rules_response = self.api.get_rules_from_device_id(
                    device_id, uid=uid, documentation=documentation
                )
                rules = self._transform_rules_response(rules_response)

                for rule in rules:
                    rule.device = device_obj

                self._device_rules_dict[device_id] = rules

                if latest_revision_id is not None:
                    self._revision_rules_dict[latest_revision_id] = rules

                if rule_uid:
                    return self._filter_rule_uid(rules, rule_uid)
                else:
                    return rules
        else:
            rules_response = self.api.get_rules_from_device_id(
                device_id, uid=uid, documentation=documentation
            )
            rules = self._transform_rules_response(rules_response)

            for rule in rules:
                rule.device = device_obj

            return rules

    def _get_nat_rules_by_device(self, device: Union[str, int]):
        def _get_response(device_obj, interface_name=None):
            device_id = device_obj.id

            nat_rules_response = self.api.get_nat_rules_from_device_id(
                device_id, input_interface=interface_name
            )
            if not nat_rules_response.ok:
                try:
                    msg = nat_rules_response.json().get("result").get("message")
                    nat_rules_response.raise_for_status()
                except HTTPError as e:
                    raise ValueError(f"got '{msg}' from Error {e}")

            nat_rules = self._transform_nat_rules_response(nat_rules_response)

            for rule in nat_rules:
                rule.device = device_obj

            return nat_rules

        should_iterate_interfaces = False

        device_obj = self.get_device(device)

        if device_obj is None:
            raise ValueError(f"Device {device} not found")

        if device_obj.vendor in [Device.Vendor.CISCO]:
            should_iterate_interfaces = True

        device_id = device_obj.id

        interfaces = None
        rules = []
        if should_iterate_interfaces:
            interfaces = self.get_interfaces(device_id)
            for interface in interfaces:
                iface_nat_rules = _get_response(device_obj, interface.name)
                for rule in iface_nat_rules:
                    rules.append(rule)
            return rules
        else:
            nat_rules = _get_response(device_obj, None)

            for rule in nat_rules:
                rules.append(rule)
            return rules

    def get_nat_rules(self, device: Union[str, int, None] = None) -> List[SecurityRule]:
        if device is None:
            raise NotImplementedError(
                "Current SDK does not support NAT rules for all devices"
            )
        elif device is not None:
            device_obj = self.get_device(device)
            to_return_nat_rules = []
            nat_rules = self._get_nat_rules_by_device(device=device)
            nat_rule_map = {}
            for n in nat_rules:
                nat_rule_map[n.id] = n

            for nat_rule in nat_rule_map.values():
                nat_rule.device = device_obj
                to_return_nat_rules.append(nat_rule)
            return to_return_nat_rules

    def get_rules(
        self,
        device: Union[str, int, None] = None,
        revision: Union[int, None] = None,
        rule_uid: Optional[str] = None,
        uid: Optional[str] = None,
        documentation: bool = True,
        cache: bool = True,
    ) -> List[SecurityRule]:
        match_rules = []
        if device is None and revision is None:
            for device in self.get_devices():
                rules = self._get_rules_by_device(
                    device=device.id,
                    rule_uid=rule_uid,
                    uid=uid,
                    documentation=documentation,
                    cache=cache,
                )
                for rule in rules:
                    rule.device = device

                    if rule_uid is not None and uids_match(rule.uid, rule_uid):
                        match_rules.append(rule)
                    elif rule_uid is None:
                        match_rules.append(rule)
        elif device is not None and revision is not None:
            raise ValueError(
                "You cannot specify both revision and device arguments for the same call"
            )

        elif device is not None:
            rules = self._get_rules_by_device(
                device=device,
                rule_uid=rule_uid,
                uid=uid,
                documentation=documentation,
                cache=cache,
            )
            device_obj = self.get_device(device)
            for rule in rules:
                rule.device = device_obj
                match_rules.append(rule)

        elif revision is not None:
            rules = self._get_rules_by_revision(
                revision=revision,
                rule_uid=rule_uid,
                uid=uid,
                documentation=documentation,
                cache=cache,
            )
            for rule in rules:
                match_rules.append(rule)
        return match_rules

    def get_rule_documentation(
        self: "St", device: Union[str, int], rule: Union[int, SecurityRule]
    ) -> Documentation:
        device_obj = self.get_device(device)
        rule_id = rule.id if isinstance(rule, SecurityRule) else rule
        r = self.api.session.get(
            f"devices/{device_obj.id}/rules/{rule_id}/documentation"
        )
        if not r.ok:
            r.raise_for_status()  # no detail msg in response
        return Documentation.kwargify(r.json()["rule_documentation"])

    def update_rule_documentation(
        self,
        device: Union[str, int],
        rule: Union[int, SecurityRule],
        rule_documentation: Documentation,
    ) -> None:
        if isinstance(device, str):
            device_obj = self.get_device(device)
            device = device_obj.id
        rule_id = rule.id if isinstance(rule, SecurityRule) else rule
        documentation_body = {"rule_documentation": rule_documentation._json}

        r = self.api.session.put(
            f"devices/{device}/rules/{rule_id}/documentation", json=documentation_body
        )

        if not r.ok:
            r.raise_for_status()

    def _prime_revisions_cache(self):
        self._revisions_dict = {}
        self._device_revisions_dict = {}

    def _get_revision_from_cache(self, revision_id: int):
        revision = self._revisions_dict.get(revision_id, None)
        if not revision:
            raise TypeError("No revision found in cache")
        return revision

    def _get_revision_from_server(self, revision_id: int):
        revision_response = self.api.session.get(f"revisions/{revision_id}")
        if not revision_response.ok:
            try:
                msg = (
                    revision_response.json().get("result").get("message")
                    if revision_response.text
                    else f"Generic API Error {revision_response.status_code}"
                )
                revision_response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"got '{msg}' from Error :{e}")

        revision_json = revision_response.json()
        revision_json = get_api_node(revision_json, "revision")
        return Revision.kwargify(revision_json)

    def get_revision(self, revision: int, cache: bool = True) -> Revision:
        if cache and self.cache:
            if not self._revisions_dict:
                self._prime_revisions_cache()
            try:
                revision_obj = self._get_revision_from_cache(revision)
            except TypeError as e:
                try:
                    revision_obj = self._get_revision_from_server(revision)
                except HTTPError as e:
                    raise ValueError(f"got error :{e}")
            self._revisions_dict[revision] = revision_obj
            return revision_obj
        else:
            return self._get_revision_from_server(revision)

    def get_latest_revision(self, device: Union[str, int]):
        device_id = self.get_device(device).id if isinstance(device, str) else device
        revision_response = self.api.session.get(f"devices/{device_id}/latest_revision")
        # API return can not be JSON Decoded - use generic error
        if not revision_response.ok:
            revision_response.raise_for_status()

        revision_json = revision_response.json()
        revision_json = get_api_node(revision_json, "revision")

        revision_obj = Revision.kwargify(revision_json)
        return revision_obj

    def _get_revisions_from_cache(self, device_id: int):
        revisions = self._device_revisions_dict.get(device_id, None)
        if revisions:
            return revisions
        else:
            return False

    def _get_revisions_from_server(self, device_id: int):
        revisions_response = self.api.session.get(f"devices/{device_id}/revisions")
        if not revisions_response.ok:
            revisions_response.raise_for_status()

        revisions_json = revisions_response.json()
        revisions_json = get_api_node(revisions_json, "revision", listify=True)

        revisions = [Revision.kwargify(revision) for revision in revisions_json]
        return revisions

    def get_revisions(self, device: Union[str, int], cache: bool = True):
        device_obj = self.get_device(identifier=device)
        if not device_obj:
            raise ValueError(f"Device {device} not found")
        device_id = device_obj.id

        if cache and self.cache:
            if not self._device_revisions_dict:
                self._prime_revisions_cache()

            revisions = self._get_revisions_from_cache(device_id)
            if revisions:
                return revisions
            else:
                revisions = self._get_revisions_from_server(device_id)
                self._device_revisions_dict[device_id] = revisions
                return revisions
        else:
            return self._get_revisions_from_server(device_id)

    def search_rules(
        self: "St",
        text: Optional[str] = None,
        devices: Union[Union[str, int], List[Union[str, int]]] = None,
        context: Optional[int] = None,
        # The list of known rule search parameters that have been encoded (so
        # far):
        shadowed: Optional[bool] = None,
        expiration_date: Optional[Union[Emptiness, date]] = None,
        certification_expiration_date: Optional[Union[Emptiness, date]] = None,
        comment: Optional[Emptiness] = None,
        # The unknowns (dun dun dun):
        #
        # (N.B. that search params passed by name in the type signature
        # (`shadowed`, etc.) will overwrite any params inside of
        # `search_text_params` (actually, if you try to call the function like:
        # `st.search_rules(shadowed=True, **{"shadowed": True})` it will crash),
        # so please prefer passing search query params by name if you can.)
        **search_text_params: Dict[str, Union[str, bool, List[Union[str, bool]]]],
    ) -> List[SecurityRule]:
        # This function operates in two stages:
        #
        # 1. First it calls the base "/rule_search" endpoint with your given
        # device list (or all devices if no device list was passed) and search
        # queries to see which devices will be sub-queried.
        #
        # 1. Next, it consecutively calls the "/rule_search/{device_id}"
        # endpoint per device, and returns every rule that matches your query.

        # For whatever reason, rule_search supports a format that wants you to
        # put your search text and your other queries all together in the same
        # `search_text` URI parameter. However, actual search `str` *text* must
        # be specified first. So instead of something like this:
        #
        #     ?search_text=my cool search text&shadowed=true&action=accept
        #
        # We have to do this:
        #
        #     ?search_text=my cool search text shadowed:true action:accept
        #
        # N.B.: It seems that the API parses ':' and '%3A' into the same value,
        # so it currently unknown to me how to include a literal `':'` in the
        # text portion of an programmatic hit of this endpoint (without that
        # text portion being treated as a key:value pair).
        devices_cache = {d.id: d for d in self.get_devices()}
        if devices is None:
            devices = devices_cache.keys()
        if not isinstance(devices, Iterable):
            devices = [devices]
        _search_text_params = []
        for k, v in search_text_params.items():
            if k == "uid":
                v = sanitize_uid(v)
            _search_text_params.append(_querify(k, v))
        search_text_params = _search_text_params
        search_text_string = " ".join(search_text_params)
        if expiration_date is not None:
            if isinstance(expiration_date, Emptiness):
                string = f"{expiration_date.value}"
            else:
                string = f":{expiration_date.strftime('%Y%m%d')}"
            search_text_params.append("expirationdate" + string)

        if certification_expiration_date is not None:
            if isinstance(certification_expiration_date, Emptiness):
                string = f"{certification_expiration_date.value}"
            else:
                string = f":{certification_expiration_date.strftime('%Y%m%d')}"
            search_text_params.append("certificationexpirationdate" + string)

        if comment is not None:
            search_text_params.append(f"comment{comment.value}")

        def _chunked_rule_search(devices):
            LOGGER.debug(f"Running chunked rule search for devices: {devices}")

            params = OrderedDict(
                {
                    "devices": ",".join(str(d) for d in devices),
                    "search_text": text + " " if text is not None else "",
                }
            )
            if context:
                params["context"] = context

            params["search_text"] += search_text_string

            if shadowed is not None:
                search_text_params.append(f"shadowed:{_bool(shadowed)}")

            # N.B.: It *is* possible to save one HTTP request here if we only have
            # a single device id in `devices`, as we can skip the `/rule_search`
            # hit and just request `/rule_search/{devices[0]}` immediately, but it
            # wasn't worth the complexity of implementation as `/rule_search` is
            # plenty fast.
            rule_search_info = self.api.session.get("rule_search", params=params)
            if not rule_search_info.ok:
                try:
                    msg = rule_search_info.json().get("result").get("message")
                    rule_search_info.raise_for_status()
                except HTTPError as e:
                    raise ValueError(f"got '{msg}' from API Error: {e}")

            rule_search_info = rule_search_info.json()

            # Walrus, come save me.
            rule_search_info = rule_search_info.get("device_list")
            if rule_search_info is None:
                return
            rule_search_info = rule_search_info.get("device")
            if rule_search_info is None:
                return

            # Save me twice.
            devices = [
                {
                    "rule_count": device_info["rule_count"],
                    "device_id": device_info["device_id"],
                }
                for device_info in rule_search_info
                if device_info.get("device_id") is not None
                and device_info.get("rule_count", 0) > 0
            ]

            return devices

        total_rule_count = 0

        devices = list(devices)  # Turn dict_keys() into list for chunking below
        i = 0
        found_rules = []
        while i < len(devices):
            device_ids = devices[i : i + 50]
            i += 50

            _devices_chunk = _chunked_rule_search(device_ids)

            total_rule_count += sum([d["rule_count"] for d in _devices_chunk])

            for device_info in _devices_chunk:
                device_id = device_info["device_id"]
                device_rule_count = device_info["rule_count"]

                device = devices_cache.get(device_id)
                if device is None:
                    LOGGER.warning(
                        {
                            "message": f"There is no device known to SecureTrack with id {device_id}; perhaps it was deleted? Skipping.",
                            "device": {"id": device_id},
                        }
                    )
                    continue

                # N.B.: API results are 0-based, not one-based. So passing in a
                # `count` of 3000 will return items starting from the 3001st and so
                # forth (passing in 0 will return starting from item #1).
                params = OrderedDict(
                    {"start": 0, "count": 3000, "search_text": search_text_string}
                )
                rules_retrieved_for_this_device = 0

                while rules_retrieved_for_this_device < device_rule_count:
                    rules = self.api.session.get(
                        f"rule_search/{device.id}", params=params
                    )
                    params["start"] += params["count"]

                    if not rules.ok:
                        try:
                            msg = rules.json().get("result").get("message")
                            rules.raise_for_status()
                        except HTTPError as e:
                            raise ValueError(f"got '{msg}' from API error: {e}")
                    rules = rules.json()

                    if rules.get("rules") is None:
                        break
                    if rules["rules"].get("rule") is None:
                        break
                    if rules["rules"].get("count") is None:
                        break

                    for rule in rules["rules"]["rule"]:
                        rule = SecurityRule.kwargify(rule)
                        rule.total_rule_count = total_rule_count
                        rule.device_rule_count = device_rule_count
                        rule.device = device
                        found_rules.append(rule)

                    rules_retrieved_for_this_device += rules["rules"]["count"]
        return found_rules

    def rule_search(self, *args: tuple, **kwargs: dict) -> List[SecurityRule]:
        return self.search_rules(*args, **kwargs)

    def update_documentation(
        self, device_id: int, rule_id: int, rule_doc: Documentation
    ):
        response = self.api.session.put(
            f"devices/{device_id}/rules/{rule_id}/documentation", json=rule_doc._json
        )

        if not response.ok:
            response.raise_for_status()

    def get_device_policies(self, device: Union[int, str]) -> List[BindingPolicy]:
        if isinstance(device, str):
            device_obj = self.get_device(device)
            if not device_obj:
                raise ValueError(f"Cannot find device {device}")

            device_id = device_obj.id
        else:
            device_id = device

        response = self.api.session.get(f"devices/{device_id}/policies")
        if not response.ok:
            response.raise_for_status()

        _json = response.json()

        policies = get_api_node(_json, "policies", listify=True)
        to_return_policies = []
        for policy in policies:
            try:
                policy_obj = BindingPolicy.kwargify(policy)
                to_return_policies.append(policy_obj)
            except Exception as e:  # pragma: no cover
                raise ValueError(
                    f"unable to kwargify policy_json: {policy}, got error: {e}"
                )
        return to_return_policies

    def get_device_policy(self, device: Union[int, str], policy: str) -> BindingPolicy:
        policies = self.get_device_policies(device)

        for policy_obj in policies:
            if policy_obj.name == policy:
                return policy_obj

        raise ValueError("No matching policy found in given device.")

    def get_interfaces(self, device_id: int) -> List[Interface]:
        device_info = self.default.get_device(device_id)

        if device_info and device_info.vendor.name == "CHECKPOINT":
            interfaces = self.api.get_topology_interfaces_from_device_id(device_id)
            base_id = "interface"
        else:
            interfaces = self.api.get_interfaces_from_device_id(device_id)
            base_id = "interfaces.interface"

        if interfaces.status_code == 404:
            raise ValueError(f"Device {device_id} not found")
        elif interfaces.status_code == 400:
            # checkpoint devices unsupported returns 400
            raise ValueError(
                f"CheckPoint Device {device_id} not supported, use topology_interfaces"
            )
        else:
            return [
                Interface.kwargify(d)
                for d in get_api_node(interfaces.json(), base_id, listify=True)
            ]

    def get_bindable_objects(self, device_id: int) -> List[BindableObject]:
        objects = self.api.get_bindable_objects_from_device_id(device_id)
        if objects.status_code == 404:
            raise ValueError(f"Device {device_id} not found")
        else:
            return [
                BindableObject.kwargify(d)
                for d in get_api_node(objects.json(), "bindable_objects", listify=True)
            ]

    def get_topology_interfaces(
        self, device_id: int, is_generic: Optional[int] = 0
    ) -> List[TopologyInterface]:
        interfaces = self.api.get_topology_interfaces_from_device_id(
            device_id, is_generic=is_generic
        )

        return [
            TopologyInterface.kwargify(d)
            for d in get_api_node(interfaces.json(), "interface", listify=True)
        ]

    def get_generic_devices(
        self,
        name: Optional[str] = None,
        context: Optional[int] = None,
        cache: bool = True,
    ) -> List[GenericDevice]:
        if not cache:
            response = self.api.get_generic_devices(name=name, context=context)
            if not response.ok:
                try:
                    msg = response.text
                    response.raise_for_status()
                except HTTPError as e:
                    raise ValueError(
                        f"Got {e}, with Error Message: {msg} from generic_devices API"
                    )
            else:
                generic_devices = get_api_node(
                    response.json(), "generic_devices.device", listify=True
                )
                return [GenericDevice.kwargify(d) for d in generic_devices]
        else:
            if self._generic_devices_cache.is_empty():
                self._prime_generic_devices_cache()

            if name is None:
                return self._generic_devices_cache.get_data()
            else:
                for d in self._generic_devices_cache.get_data():
                    if name == d.name:
                        return d

    def delete_generic_device(
        self, identifier: Union[int, str], update_topology: bool = False
    ) -> None:
        if isinstance(identifier, str):
            if self._generic_devices_cache.is_empty():
                self._prime_generic_devices_cache()

            device = self._generic_devices_index.get(identifier)
            if device is None:
                raise ValueError("Could not find device with specified name")

            identifier = device.id

        response = self.api.delete_generic_device(
            id=identifier, update_topology=update_topology
        )

        if not response.ok:
            try:
                msg = response.text
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"got '{msg}' from API Error: {e}")

    def add_generic_device(
        self,
        name: str,
        configuration: Union[BytesIO, str],
        update_topology: bool = False,
        customer_id: Optional[int] = None,
    ) -> None:
        response = self.api.post_generic_device(
            name=name,
            configuration=configuration,
            update_topology=update_topology,
            customer_id=customer_id,
        )

        if not response.ok:
            try:
                msg = response.text
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"got '{msg}' from API Error: {e}")

    def update_generic_device(
        self,
        id: int,
        name: str,
        configuration: Union[BytesIO, str],
        update_topology: bool = False,
    ) -> None:
        response = self.api.put_generic_device(
            id=id,
            configuration=configuration,
            name=name,
            update_topology=update_topology,
        )

        if not response.ok:
            try:
                msg = response.text
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"got '{msg}' from API Error :{e}")

    def import_generic_device(
        self,
        name: str,
        configuration: Union[BytesIO, str],
        update_topology: bool = False,
        customer_id: Optional[int] = None,
    ):
        if self._generic_devices_cache.is_empty():
            self._prime_generic_devices_cache()

        existing_device = self._generic_devices_index.get(name)

        if existing_device is None:
            return self.add_generic_device(
                name=name,
                configuration=configuration,
                update_topology=update_topology,
                customer_id=customer_id,
            )
        else:
            return self.update_generic_device(
                id=existing_device.id,
                name=name,
                configuration=configuration,
                update_topology=False,
            )

    def sync_topology(self, full_sync: bool = False):
        response = self.api.session.post(
            "topology/synchronize", params={"full_sync": full_sync}
        )

        if response.ok:
            return
        elif response.status_code == 401:
            raise ValueError("Authentication error")
        elif response.status_code == 500:
            raise ValueError("Error synchronizing topology model")
        else:
            response.raise_for_status()

    def get_topology_sync_status(self) -> TopologySyncStatus:
        response = self.api.session.get("topology/synchronize/status")

        if response.ok:
            status = get_api_node(response.json(), "status")
            status = TopologySyncStatus.kwargify(status)
            return status
        elif response.status_code == 401:
            raise ValueError("Authentication error")
        elif response.status_code == 500:
            raise ValueError("Error getting synchronization process")
        else:
            response.raise_for_status()

    def _generic_body(self, type: str, data: Union[dict, List[dict]]) -> dict:
        return {type: data if isinstance(data, typing.List) else [data]}

    def add_generic_interface(self, interface: dict) -> None:
        data = self._generic_body("GenericInterfaces", interface)
        response = self.api.session.post("topology/generic/interface", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Interface: {e}")

    def add_generic_interfaces(self, interfaces: List[dict]) -> None:
        data = self._generic_body("GenericInterfaces", interfaces)
        response = self.api.session.post("topology/generic/interface", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Interfaces: {e}")

    def get_generic_interface(self, int_id: Union[int, str]) -> GenericInterface:
        response = self.api.session.get(f"topology/generic/interface/{int_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Interface: {e}")

        interface = get_api_node(response.json(), "GenericInterface")
        return GenericInterface.kwargify(interface)

    def get_generic_interfaces(
        self, mgmt_id: Union[int, str]
    ) -> List[GenericInterface]:
        response = self.api.session.get(f"topology/generic/interface/mgmt/{mgmt_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Interfaces: {e}")

        return [
            GenericInterface.kwargify(d)
            for d in get_api_node(response.json(), "GenericInterfaces", listify=True)
        ]

    def update_generic_interface(self, interface: dict) -> None:
        data = self._generic_body("GenericInterfaces", interface)
        response = self.api.session.put("topology/generic/interface", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Interface: {e}")

    def update_generic_interfaces(self, interfaces: List[dict]) -> None:
        data = self._generic_body("GenericInterfaces", interfaces)
        response = self.api.session.put("topology/generic/interface", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Interfaces: {e}")

    def delete_generic_interface(self, int_id: Union[int, str]) -> None:
        response = self.api.session.delete(f"topology/generic/interface/{int_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Interface: {e}")

    def delete_generic_interfaces(self, mgmt_id: Union[int, str]) -> None:
        response = self.api.session.delete(f"topology/generic/interface/mgmt/{mgmt_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Interfaces: {e}")

    def add_generic_route(self, route: dict) -> None:
        data = self._generic_body("GenericRoutes", route)
        response = self.api.session.post("topology/generic/route", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Route: {e}")

    def add_generic_routes(self, routes: List[dict]) -> None:
        data = self._generic_body("GenericRoutes", routes)
        response = self.api.session.post("topology/generic/route", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Routes: {e}")

    def get_generic_route(self, int_id: Union[int, str]) -> GenericRoute:
        response = self.api.session.get(f"topology/generic/route/{int_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Route: {e}")

        route = get_api_node(response.json(), "GenericRoute")
        return GenericRoute.kwargify(route)

    def get_generic_routes(self, mgmt_id: Union[int, str]) -> List[GenericRoute]:
        response = self.api.session.get(f"topology/generic/route/mgmt/{mgmt_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Routes: {e}")

        return [
            GenericRoute.kwargify(d)
            for d in get_api_node(response.json(), "GenericRoutes", listify=True)
        ]

    def update_generic_route(self, route: dict) -> None:
        data = self._generic_body("GenericRoutes", route)
        response = self.api.session.put("topology/generic/route", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Route: {e}")

    def update_generic_routes(self, routes: List[dict]) -> None:
        data = self._generic_body("GenericRoutes", routes)
        response = self.api.session.put("topology/generic/route", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Routes: {e}")

    def delete_generic_route(self, int_id: Union[int, str]) -> None:
        response = self.api.session.delete(f"topology/generic/route/{int_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Route: {e}")

    def delete_generic_routes(self, mgmt_id: Union[int, str]) -> None:
        response = self.api.session.delete(f"topology/generic/route/mgmt/{mgmt_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Routes: {e}")

    def add_generic_vpn(self, vpn: dict) -> None:
        data = self._generic_body("GenericVpns", vpn)
        response = self.api.session.post("topology/generic/vpn", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Vpn: {e}")

    def add_generic_vpns(self, vpns: List[dict]) -> None:
        data = self._generic_body("GenericVpns", vpns)
        response = self.api.session.post("topology/generic/vpn", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Vpns: {e}")

    def get_generic_vpn(self, int_id: Union[int, str]) -> GenericVpn:
        response = self.api.session.get(f"topology/generic/vpn/{int_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Vpn: {e}")

        vpn = get_api_node(response.json(), "GenericVpn")
        return GenericVpn.kwargify(vpn)

    def get_generic_vpns(self, device_id: Union[int, str]) -> List[GenericVpn]:
        response = self.api.session.get(f"topology/generic/vpn/device/{device_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Vpns: {e}")

        return [
            GenericVpn.kwargify(d)
            for d in get_api_node(response.json(), "GenericVpns", listify=True)
        ]

    def update_generic_vpn(self, vpn: dict) -> None:
        data = self._generic_body("GenericVpns", vpn)
        response = self.api.session.put("topology/generic/vpn", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Vpn: {e}")

    def update_generic_vpns(self, vpns: List[dict]) -> None:
        data = self._generic_body("GenericVpns", vpns)
        response = self.api.session.put("topology/generic/vpn", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Vpns: {e}")

    def delete_generic_vpn(self, int_id: Union[int, str]) -> None:
        response = self.api.session.delete(f"topology/generic/vpn/{int_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Vpn: {e}")

    def delete_generic_vpns(self, device_id: Union[int, str]) -> None:
        response = self.api.session.delete(f"topology/generic/vpn/device/{device_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Vpns: {e}")

    def add_generic_transparent_firewalls(self, firewalls: List[dict]) -> None:
        data = self._generic_body("TransparentFirewalls", firewalls)
        response = self.api.session.post("topology/generic/transparentfw", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Transparent Firewalls: {e}")

    def get_generic_transparent_firewalls(
        self, device_id: Union[int, str]
    ) -> List[GenericTransparentFirewall]:
        response = self.api.session.get(
            f"topology/generic/transparentfw/device/{device_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Transparent Firewalls: {e}")

        return [
            GenericTransparentFirewall.kwargify(d)
            for d in get_api_node(response.json(), "TransparentFirewalls", listify=True)
        ]

    def update_generic_transparent_firewalls(self, firewalls: List[dict]) -> None:
        data = self._generic_body("TransparentFirewalls", firewalls)
        response = self.api.session.put("topology/generic/transparentfw", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Transparent Firewalls: {e}")

    def delete_generic_transparent_firewall(
        self, layer_2_data_id: Union[int, str]
    ) -> None:
        response = self.api.session.delete(
            f"topology/generic/transparentfw/{layer_2_data_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Transparent Firewall: {e}")

    def delete_generic_transparent_firewalls(self, device_id: Union[int, str]) -> None:
        response = self.api.session.delete(
            f"topology/generic/transparentfw/device/{device_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Transparent Firewalls: {e}")

    def add_generic_ignored_interfaces(self, interfaces: List[dict]) -> None:
        data = self._generic_body("IgnoredInterfaces", interfaces)
        response = self.api.session.post("topology/generic/ignoredinterface", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Ignored Interfaces: {e}")

    def get_generic_ignored_interfaces(
        self, mgmt_id: Union[int, str]
    ) -> List[GenericIgnoredInterface]:
        response = self.api.session.get(
            f"topology/generic/ignoredinterface/mgmt/{mgmt_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Ignored Interfaces: {e}")

        return [
            GenericIgnoredInterface.kwargify(d)
            for d in get_api_node(response.json(), "IgnoredInterfaces", listify=True)
        ]

    def delete_generic_ignored_interfaces(self, mgmt_id: Union[int, str]) -> None:
        response = self.api.session.delete(
            f"topology/generic/ignoredinterface/mgmt/{mgmt_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Ignored Interfaces: {e}")

    def add_generic_interface_customer(self, interface_customer: dict) -> None:
        data = self._generic_body("InterfaceCustomerTag", interface_customer)
        response = self.api.session.post(
            "topology/generic/interfacecustomer", json=data
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Interface Customer Tag: {e}")

    def add_generic_interface_customers(self, interface_customers: List[dict]) -> None:
        data = self._generic_body("InterfaceCustomerTags", interface_customers)
        response = self.api.session.post(
            "topology/generic/interfacecustomer", json=data
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Generic Interface Customer Tags: {e}")

    def get_generic_interface_customer(
        self, int_cust_id: Union[int, str]
    ) -> GenericInterfaceCustomer:
        response = self.api.session.get(
            f"topology/generic/interfacecustomer/{int_cust_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Interface Customer Tag: {e}")

        interface_customer = get_api_node(response.json(), "InterfaceCustomerTag")
        return GenericInterfaceCustomer.kwargify(interface_customer)

    def get_generic_interface_customers(
        self, device_id: Union[int, str]
    ) -> List[GenericInterfaceCustomer]:
        response = self.api.session.get(
            f"topology/generic/interfacecustomer/device/{device_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Generic Interface Customer Tags: {e}")

        return [
            GenericInterfaceCustomer.kwargify(d)
            for d in get_api_node(
                response.json(), "InterfaceCustomerTags", listify=True
            )
        ]

    def update_generic_interface_customer(self, interface_customer: dict) -> None:
        data = self._generic_body("InterfaceCustomerTag", interface_customer)
        response = self.api.session.put("topology/generic/interfacecustomer", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Interface Customer Tag: {e}")

    def update_generic_interface_customers(
        self, interface_customers: List[dict]
    ) -> None:
        data = self._generic_body("InterfaceCustomerTags", interface_customers)
        response = self.api.session.put("topology/generic/interfacecustomer", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Generic Interface Customer Tag: {e}")

    def delete_generic_interface_customer(self, int_cust_id: Union[int, str]) -> None:
        response = self.api.session.delete(
            f"topology/generic/interfacecustomer/{int_cust_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Interface Customer Tag: {e}")

    def delete_generic_interface_customers(self, device_id: Union[int, str]) -> None:
        response = self.api.session.delete(
            f"topology/generic/interfacecustomer/device/{device_id}"
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Generic Interface Customer Tags: {e}")

    def add_join_cloud(self, join_cloud: dict) -> None:
        data = self._generic_body("JoinCloud", join_cloud)
        response = self.api.session.post("topology/join/clouds", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Join Cloud: {e}")

    def get_join_cloud(self, cloud_id: Union[int, str]) -> JoinCloud:
        response = self.api.session.get(f"topology/join/clouds/{cloud_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Join Cloud: {e}")
        return JoinCloud.kwargify(response.json())

    def update_join_cloud(self, join_cloud: dict) -> None:
        data = self._generic_body("JoinCloud", join_cloud)
        response = self.api.session.put("topology/join/clouds", json=data)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Updating Join Cloud: {e}")

    def delete_join_cloud(self, cloud_id: Union[int, str]) -> None:
        response = self.api.session.delete(f"topology/join/clouds/{cloud_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Join Cloud: {e}")



import pytest
import json
import responses

from pytos2.securetrack.generic_route import GenericRoute
from pytos2.utils import get_api_node


class TestGenericRoute:
    mgmt = json.load(open("tests/securetrack/json/generic_routes/mgmt-3.json"))
    routes = mgmt["GenericRoutes"]

    route27 = json.load(open("tests/securetrack/json/generic_routes/route-27.json"))
    route = route27["GenericRoute"]

    @responses.activate
    def test_add_generic_routes_200(self, st):
        """Add one or multiple routes"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=201,
        )

        """Add single route"""
        newRoute = st.add_generic_route(self.route)
        assert newRoute == None

        """Add multiple routes"""
        newRoutes = st.add_generic_routes(self.routes)
        assert newRoutes == None

    @responses.activate
    def test_add_generic_routes_400(self, st):
        """Add one or multiple routes"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=400,
        )

        """Add single route"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_route(self.route)
        assert "Bad Request" in str(exception.value)

        """Add multiple routes"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_routes(self.routes)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_routes_404(self, st):
        """Add one or multiple routes"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=404,
        )

        """Add single route"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_route(self.route)
        assert "Not Found" in str(exception.value)

        """Add multiple routes"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_routes(self.routes)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_route(self, st, generic_route_mock):
        """Get route by route id"""
        mockRes = get_api_node(self.route27, "GenericRoute")
        route = GenericRoute.kwargify(mockRes)
        routeByInt = st.get_generic_route(27)
        routeByStr = st.get_generic_route("27")
        assert routeByInt == route
        assert routeByStr == route

        with pytest.raises(ValueError) as exception:
            st.get_generic_route(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_route("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_routes(self, st, generic_route_mock):
        """Get routes by mgmt id"""
        routes = [
            GenericRoute.kwargify(d)
            for d in get_api_node(self.mgmt, "GenericRoutes", listify=True)
        ]
        routesByInt = st.get_generic_routes(3)
        routesByStr = st.get_generic_routes("3")
        assert routesByInt == routes
        assert routesByStr == routes

        with pytest.raises(ValueError) as exception:
            st.get_generic_routes(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_routes("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_routes_200(self, st):
        """Update one or multiple routes"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=200,
        )

        """Update single route"""
        newRoute = st.update_generic_route(self.route)
        assert newRoute == None

        """Update multiple routes"""
        newRoutes = st.update_generic_routes(self.routes)
        assert newRoutes == None

    @responses.activate
    def test_update_generic_routes_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_route(self.route)
        assert "Bad Request" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_routes(self.routes)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_routes_404(self, st):
        """PUT mgmt id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/route",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            st.update_generic_route(self.route)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.update_generic_routes(self.routes)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_route(self, st, generic_route_mock):
        """Delete route by route id"""
        routeByInt = st.delete_generic_route(27)
        routeByStr = st.delete_generic_route("27")
        assert routeByInt == None
        assert routeByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_route(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_route("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_routes(self, st, generic_route_mock):
        """Delete routes by mgmt id"""
        routesByInt = st.delete_generic_routes(2)
        routesByStr = st.delete_generic_routes("2")
        assert routesByInt == None
        assert routesByStr == None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_routes(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_generic_routes("404")
        assert "Not Found" in str(exception.value)



from enum import Enum
from typing import Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop

from netaddr import IPAddress


@propify
class GenericRoute(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int))
    mgmtId: Optional[int] = prop(None, converter=optional(int))
    destination: Optional[IPAddress] = prop(None)
    mask: Optional[IPAddress] = prop(None, repr=False)
    interfaceName: str = prop(None)
    nextHop: Optional[str] = prop(None)
    nextHopType: Optional[str] = prop(None)
    vrf: Optional[str] = prop(None)



from enum import Enum
from typing import Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop

from netaddr import IPAddress


@propify
class GenericInterface(Jsonable):
    class Type(Enum):
        EXTERNAL = "external"
        INTERNAL = "internal"

    id: Optional[int] = prop(None, converter=optional(int))
    mgmtId: Optional[int] = prop(None, converter=optional(int))
    name: str = prop(None)
    ip: Optional[IPAddress] = prop(None)
    mask: Optional[IPAddress] = prop(None, repr=False)
    vrf: Optional[str] = prop(None)
    mpls: Optional[bool] = prop(False)
    unnumbered: Optional[bool] = prop(False)
    type: Optional[Type] = prop(None)



import pytest
import json
import responses

from netaddr import IPAddress

from pytos2.securetrack.device import Device
from pytos2.utils import get_api_node


class TestDevice:
    @pytest.fixture
    def device(self):
        j = json.load(open("tests/securetrack/json/devices/devices.json"))
        device_node = get_api_node(j, "devices.device")[0]
        return Device.kwargify(device_node)

    def test_attributes(self, device):
        assert device.id == 1
        assert device.name == "RTR1"
        assert device.vendor == Device.Vendor.CISCO
        assert device.model == Device.Model.ROUTER
        assert isinstance(device.domain_id, int) and device.domain_id == 1
        assert device.domain_name == "Default"
        assert device.module_uid == ""
        assert device.ip == IPAddress("10.100.200.54")
        assert (
            isinstance(device.latest_revision, int) and device.latest_revision == 1674
        )
        assert device.virtual_type == ""

    def test_set_attributes(self, device):
        assert device.ip == IPAddress("10.100.200.54")

        device.ip = "1.2.3.4"
        assert device.ip == IPAddress("1.2.3.4")

    @responses.activate
    def test_properties(self, devices_mock, st):
        device = st.get_device(identifier=60)
        assert device
        assert device.name == "NSX-Edge-01"

        parent_device = device.parent
        assert parent_device.id == 58
        assert parent_device.name == "NSX"

        assert len(parent_device.children) == 4

        grandparent = parent_device.parent
        assert grandparent is None



import json
from pathlib import Path

from netaddr import IPAddress
import responses
import pytest
from netaddr import IPAddress, IPNetwork, IPRange

from pytos2.securetrack.network_object import classify_network_object
from pytos2.utils import get_api_node


class TestDevice:
    @pytest.fixture
    def device(self):
        j = json.load(open("tests/securetrack/json/devices/devices.json"))
        device_node = get_api_node(j, "devices.device")[0]
        return classify_network_object(device_node)


class TestLoadNetworkObjects:
    def test_all_objects(self):
        for path in Path("tests/securetrack/json/network_objects/").glob("*.json"):
            device_id = str(path).split("/")[-1].split(".")[0]
            if device_id.isdigit():
                network_objects_node = get_api_node(
                    json.load(path.open()), "network_objects.network_object"
                )
                for network_object in network_objects_node:
                    classify_network_object(network_object)


class TestCheckpointObject:
    @responses.activate
    def test_nat_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(
            device=20, name="CP_default_Office_Mode_addresses_pool"
        )
        assert obj.nat_info.mapped_to_ip == "Hide Behind Gateway"
        obj = st.get_network_object(device=20, name="Net_172.16.40.0")
        assert isinstance(obj.nat_info.mapped_to_ip, IPAddress)


class TestFortigateObject:
    @responses.activate
    def test_nat_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=157, name="VIP to HQ")
        assert isinstance(obj.nat_info.mapped_ip, IPAddress)
        assert isinstance(obj.nat_info.mapped_ip_max, IPAddress)


class TestHostObject:
    # this class covers host object with interfaces as well
    @responses.activate
    def test_host_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=184, name="SGW_200.237")
        assert isinstance(obj.ip, IPAddress)
        assert isinstance(obj.interfaces[0].interface_ips[0].ip, IPAddress)


class TestSubnetObject:
    @responses.activate
    def test_subnet_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=184, name="Sales_1")
        assert isinstance(obj.subnet, IPNetwork)


class TestRangeObject:
    @responses.activate
    def test_range_object(self, st, network_objects_mock):
        # this should be refactored when get_network_object returns an object
        obj = st.get_network_object(device=184, name="my_range")
        assert isinstance(obj.range, IPRange)
